using System;
using System.Text;
using System.Reflection;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Crypto.Asymmetric;

using bcfipsin100.util;

public class DHExamples
{

    public static void Main(String[] args)
    {
                
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("=================================");
        Console.WriteLine("Start Of DHExamples.Main");
        Console.WriteLine("=================================");

        DHDomainParameters ffdhe3072 = selectParameters(FipsDH.DomainParams.ffdhe3072);
        string  pParam = ffdhe3072.P.ToString();       
        string  qParam = ffdhe3072.Q.ToString();       
        string  gParam = ffdhe3072.G.ToString();       
        Console.WriteLine(string.Format("DH Params ffdhe3072:\tp=[{0}]\tq=[{1}]\tg=[{2}]",
            pParam.Substring(0,20)+"..." + pParam.Substring(pParam.Length-20,20),
            qParam.Substring(0,20)+"..." + qParam.Substring(pParam.Length-20,20),
            gParam));

        DHDomainParameters modp3072 = selectParameters(FipsDH.DomainParams.modp3072);
        pParam = modp3072.P.ToString();       
        qParam = modp3072.Q.ToString();       
        gParam = modp3072.G.ToString();       
        Console.WriteLine(string.Format("DH Params modp3072:\tp=[{0}]\tq=[{1}]\tg=[{2}]",
            pParam.Substring(0,20)+"..." + pParam.Substring(pParam.Length-20,20),
            qParam.Substring(0,20)+"..." + qParam.Substring(pParam.Length-20,20),
            gParam));
            
        DHDomainParameters dhDomain = generateParameters(2048,224, ExValues.DHFixedRandomA(1000));
        pParam = dhDomain.P.ToString(16);       
        qParam = dhDomain.Q.ToString(16);       
        gParam = dhDomain.G.ToString(16);       
        Console.WriteLine(string.Format("DH Params   DSA:\tp=[{0}]\tq=[{1}]\tg=[{2}]",
            pParam.Substring(0,20)+"..." + pParam.Substring(pParam.Length-20,20),
            qParam,
            gParam.Substring(0,20)+"..." + pParam.Substring(pParam.Length-20,20)));
            
        dhDomain = generateParameters(dhDomain.P,dhDomain.Q,dhDomain.G);
        pParam = dhDomain.P.ToString(16);       
        qParam = dhDomain.Q.ToString(16);       
        gParam = dhDomain.G.ToString(16);       
        Console.WriteLine(string.Format("DH Params  User:\tp=[{0}]\tq=[{1}]\tg=[{2}]",
            pParam.Substring(0,20)+"..." + pParam.Substring(pParam.Length-20,20),
            qParam,
            gParam.Substring(0,20)+"..." + pParam.Substring(pParam.Length-20,20)));

        AsymmetricKeyPair<AsymmetricDHPublicKey, AsymmetricDHPrivateKey> keyPairInitiator = 
            generateDHKeyPairUsingParameters(dhDomain,ExValues.dhKeyAFixedRandom);  
        string priKeyString = keyPairInitiator.PrivateKey.X.ToString();
        string pubKeyString = keyPairInitiator.PublicKey.Y.ToString();
        Console.WriteLine(string.Format("DH Keys Initiator:\tPri=[{0}]\tPub=[{1}]",
            priKeyString.Substring(0,20) + "..." + priKeyString.Substring(priKeyString.Length-20,20),
            pubKeyString.Substring(0,20) + "..." + pubKeyString.Substring(pubKeyString.Length-20,20)
            ));
            
        AsymmetricKeyPair<AsymmetricDHPublicKey, AsymmetricDHPrivateKey> keyPairRecipient = 
            generateDHKeyPairUsingParameters(dhDomain,ExValues.dhKeyBFixedRandom);  
        priKeyString = keyPairRecipient.PrivateKey.X.ToString();
        pubKeyString = keyPairRecipient.PublicKey.Y.ToString();
        Console.WriteLine(string.Format("DH Keys Recipient:\tPri=[{0}]\tPub=[{1}]",
            priKeyString.Substring(0,20) + "..." + priKeyString.Substring(priKeyString.Length-20,20),
            pubKeyString.Substring(0,20) + "..." + pubKeyString.Substring(pubKeyString.Length-20,20)
            ));

        
        byte[] initAgreement = initiatorAgreementBasic(keyPairInitiator.PrivateKey, keyPairRecipient.PublicKey);
        Console.WriteLine(string.Format("Initiator Basic Agreement:\tKey=[{0}]\tLen=[{1}]",
            string.Join("",ExValues.SubArray(initAgreement,0,20)) + "..." + 
            string.Join("",ExValues.SubArray(initAgreement,initAgreement.Length-20,20)),
            initAgreement.Length
            ));
        byte[] recpAgreement = recipientAgreementBasic(keyPairRecipient.PrivateKey, keyPairInitiator.PublicKey);
        Console.WriteLine(string.Format("Recipient Basic Agreement:\tKey=[{0}]\tLen=[{1}]",
            string.Join("",ExValues.SubArray(recpAgreement,0,20)) + "..." + 
            string.Join("",ExValues.SubArray(recpAgreement,recpAgreement.Length-20,20)),
            recpAgreement.Length
            ));

        initAgreement = initiatorAgreementWithConcatKDF(keyPairInitiator.PrivateKey, keyPairRecipient.PublicKey,
                                    ExValues.additionalAuthenticatedDataA);
        Console.WriteLine(string.Format("Initiator Basic Agreement:\tKey=[{0}]\tLen=[{1}]",
            string.Join("",initAgreement), initAgreement.Length));
        recpAgreement = recipientAgreementWithConcatKDF(keyPairRecipient.PrivateKey, keyPairInitiator.PublicKey,
                                    ExValues.additionalAuthenticatedDataA);
        Console.WriteLine(string.Format("Recipient Basic Agreement:\tKey=[{0}]\tLen=[{1}]",
            string.Join("",recpAgreement), recpAgreement.Length));
        

        initAgreement = initiatorAgreementWithX963KDF(keyPairInitiator.PrivateKey, keyPairRecipient.PublicKey,
                                    ExValues.additionalAuthenticatedDataA);
        Console.WriteLine(string.Format("Initiator Basic Agreement:\tKey=[{0}]\tLen=[{1}]",
            string.Join("",initAgreement), initAgreement.Length));
        recpAgreement = recipientAgreementWithX963KDF(keyPairRecipient.PrivateKey, keyPairInitiator.PublicKey,
                                    ExValues.additionalAuthenticatedDataA);
        Console.WriteLine(string.Format("Recipient Basic Agreement:\tKey=[{0}]\tLen=[{1}]",
            string.Join("",recpAgreement), recpAgreement.Length));
        
        
        Console.WriteLine("=================================");
        Console.WriteLine("End   Of DHExamples.Main");
        Console.WriteLine("=================================");
        Console.WriteLine();
        Console.WriteLine();        
    }
   

    /// <summary>
    /// Example 38 – DH Domain Parameters Selection
    /// </summary>    
    public static DHDomainParameters selectParameters(IDHDomainParametersID dhParamsID)  
    {   
        DHDomainParameters dhParams = 
            DHDomainParametersIndex.LookupDomainParameters(dhParamsID);
        return dhParams;
    }

    /// <summary>
    /// Example 39 – DH Domain Parameters Generation
    /// </summary>    
    public static DHDomainParameters generateParameters(int L, int N, SecureRandom sRandom)  
    {
        FipsDsa.DomainGenParameters dsaDomainGenParams = new FipsDsa.DomainGenParameters(L,N);
        FipsDsa.DomainParametersGenerator dsaDomainParamsGen = 
            CryptoServicesRegistrar.CreateGenerator(dsaDomainGenParams,sRandom);    
        DsaDomainParameters dsaDomainParams = dsaDomainParamsGen.GenerateDomainParameters();

        DHDomainParameters dhParams = 
            new DHDomainParameters(dsaDomainParams.P, dsaDomainParams.Q, dsaDomainParams.G);
        return dhParams;
    }
    
    /// <summary>
    /// Example 40 – DH Domain Parameters Generation
    /// </summary>    
    public static DHDomainParameters generateParameters(BigInteger dhP,BigInteger dhQ, BigInteger dhG)  
    {   
        DHDomainParameters dhParams = new DHDomainParameters(dhP, dhQ, dhG);
        return dhParams;
    }
 
    /// <summary>
    /// Example 41 – DH Key Pair Generation
    /// </summary>    
    public static AsymmetricKeyPair<AsymmetricDHPublicKey, AsymmetricDHPrivateKey> 
                        generateDHKeyPairUsingParameters(
                        DHDomainParameters dhParams, SecureRandom keyFixedRandom)  
    {   
        FipsDH.KeyGenerationParameters keyGenParameters =
            new FipsDH.KeyGenerationParameters(dhParams);
                
        FipsDH.KeyPairGenerator kpGen = 
            CryptoServicesRegistrar.CreateGenerator(keyGenParameters, keyFixedRandom);

        return kpGen.GenerateKeyPair();
    }   

    /// <summary>
    /// Example 42 – Basic DH Key Agreement
    /// </summary>    
    public static byte[] initiatorAgreementBasic(
        AsymmetricDHPrivateKey initiatorPrivate, AsymmetricDHPublicKey recipientPublic)
    {
        IAgreementCalculatorService agreeFact =
            CryptoServicesRegistrar.CreateService(initiatorPrivate);

        IAgreementCalculator<FipsDH.AgreementParameters> agreement =
            agreeFact.CreateAgreementCalculator(FipsDH.DH);

        return agreement.Calculate(recipientPublic);    
    }
    
    public static byte[] recipientAgreementBasic(
        AsymmetricDHPrivateKey recipientPrivate, AsymmetricDHPublicKey initiatorPublic)
    {
        IAgreementCalculatorService agreeFact =
            CryptoServicesRegistrar.CreateService(recipientPrivate);

        IAgreementCalculator<FipsDH.AgreementParameters> agreement =
            agreeFact.CreateAgreementCalculator(FipsDH.DH);

        return agreement.Calculate(initiatorPublic);    
    }

    /// <summary>
    /// Example 43 – DH Key Agreement with Digest KDF
    /// </summary>    
    public static byte[] initiatorAgreementWithConcatKDF(
        AsymmetricDHPrivateKey initiatorPrivate, AsymmetricDHPublicKey recipientPublic, 
        byte[] userKeyingMaterial)
    {
        IAgreementCalculatorService agreeFact =
            CryptoServicesRegistrar.CreateService(initiatorPrivate);

        IAgreementCalculator<FipsDH.AgreementParameters> agreement =
            agreeFact.CreateAgreementCalculator(
                FipsDH.DH.WithKeyMaterialGenerator(
                   new FipsKdfKmg(FipsKdf.Concatenation, FipsPrfAlgorithm.Sha384, userKeyingMaterial, 32)));

        return agreement.Calculate(recipientPublic);    
    }
    
    public static byte[] recipientAgreementWithConcatKDF(
        AsymmetricDHPrivateKey recipientPrivate, AsymmetricDHPublicKey initiatorPublic,
        byte[] userKeyingMaterial)
    {
        IAgreementCalculatorService agreeFact =
            CryptoServicesRegistrar.CreateService(recipientPrivate);

        IAgreementCalculator<FipsDH.AgreementParameters> agreement =
            agreeFact.CreateAgreementCalculator(
                FipsDH.DH.WithKeyMaterialGenerator(
                   new FipsKdfKmg(FipsKdf.Concatenation, FipsPrfAlgorithm.Sha384, userKeyingMaterial, 32)));

        return agreement.Calculate(initiatorPublic);    
    }

    public static byte[] initiatorAgreementWithX963KDF(
        AsymmetricDHPrivateKey initiatorPrivate, AsymmetricDHPublicKey recipientPublic, 
        byte[] userKeyingMaterial)
    {
        IAgreementCalculatorService agreeFact =
            CryptoServicesRegistrar.CreateService(initiatorPrivate);

        IAgreementCalculator<FipsDH.AgreementParameters> agreement =
            agreeFact.CreateAgreementCalculator(
                FipsDH.DH.WithKeyMaterialGenerator(
                   new FipsKdfKmg(FipsKdf.X963, FipsPrfAlgorithm.Sha384, userKeyingMaterial, 32)));

        return agreement.Calculate(recipientPublic);    
    }
    
    public static byte[] recipientAgreementWithX963KDF(
        AsymmetricDHPrivateKey recipientPrivate, AsymmetricDHPublicKey initiatorPublic,
        byte[] userKeyingMaterial)
    {
        IAgreementCalculatorService agreeFact =
            CryptoServicesRegistrar.CreateService(recipientPrivate);

        IAgreementCalculator<FipsDH.AgreementParameters> agreement =
            agreeFact.CreateAgreementCalculator(
                FipsDH.DH.WithKeyMaterialGenerator(
                   new FipsKdfKmg(FipsKdf.X963, FipsPrfAlgorithm.Sha384, userKeyingMaterial, 32)));

        return agreement.Calculate(initiatorPublic);    
    }    
}

