using System;
using System.Text;
using System.Reflection;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Crypto.Asymmetric;

using bcfipsin100.util;

public class RsaExamples
{
    
    public static void Main(String[] args)
    {
        
        // Example of a known set of key pair        
        AsymmetricRsaPrivateKey testPrivateKey = new AsymmetricRsaPrivateKey(FipsRsa.Alg,
                new BigInteger("495370a1fb18543c16d3631e3163255df62be6eee890d5f25509e4f778a8ea6fbbbcdf85dff64e0d972003ab3681fbba6dd41fd541829b2e582de9f2a4a4e0a2d0900bef4753db3cee0ee06c7dfae8b1d53b5953218f9cceea695b08668edeaadced9463b1d790d5ebf27e9115b46cad4d9a2b8efab0561b0810344739ada0733f", 16),
                new BigInteger("010001", 16),
                new BigInteger("6c66ffe98980c38fcdeab5159898836165f4b4b817c4f6a8d486ee4ea9130fe9b9092bd136d184f95f504a607eac565846d2fdd6597a8967c7396ef95a6eeebb4578a643966dca4d8ee3de842de63279c618159c1ab54a89437b6a6120e4930afb52a4ba6ced8a4947ac64b30a3497cbe701c2d6266d517219ad0ec6d347dbe9", 16),
                new BigInteger("08dad7f11363faa623d5d6d5e8a319328d82190d7127d2846c439b0ab72619b0a43a95320e4ec34fc3a9cea876422305bd76c5ba7be9e2f410c8060645a1d29edb", 16),
                new BigInteger("0847e732376fc7900f898ea82eb2b0fc418565fdae62f7d9ec4ce2217b97990dd272db157f99f63c0dcbb9fbacdbd4c4dadb6df67756358ca4174825b48f49706d", 16),
                new BigInteger("05c2a83c124b3621a2aa57ea2c3efe035eff4560f33ddebb7adab81fce69a0c8c2edc16520dda83d59a23be867963ac65f2cc710bbcfb96ee103deb771d105fd85", 16),
                new BigInteger("04cae8aa0d9faa165c87b682ec140b8ed3b50b24594b7a3b2c220b3669bb819f984f55310a1ae7823651d4a02e99447972595139363434e5e30a7e7d241551e1b9", 16),
                new BigInteger("07d3e47bf686600b11ac283ce88dbb3f6051e8efd04680e44c171ef531b80b2b7c39fc766320e2cf15d8d99820e96ff30dc69691839c4b40d7b06e45307dc91f3f", 16));
        AsymmetricRsaPublicKey testPublicKey = new AsymmetricRsaPublicKey(FipsRsa.Alg,
                new BigInteger("495370a1fb18543c16d3631e3163255df62be6eee890d5f25509e4f778a8ea6fbbbcdf85dff64e0d972003ab3681fbba6dd41fd541829b2e582de9f2a4a4e0a2d0900bef4753db3cee0ee06c7dfae8b1d53b5953218f9cceea695b08668edeaadced9463b1d790d5ebf27e9115b46cad4d9a2b8efab0561b0810344739ada0733f", 16),
                new BigInteger("010001", 16));
                
        SymmetricSecretKey aesKey = AesExamples.generateKey();

        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("=================================");
        Console.WriteLine("Start Of RsaExamples.Main");
        Console.WriteLine("=================================");

        Console.Write("generating RSA key Pair 2048 ... ");
        AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey>  keyPair = 
            generateRSAKeyPair(BigInteger.ValueOf(0x10001), 2048);
        Console.WriteLine("done\n");
        
        // Example of extracting parameters
        /*
        Console.WriteLine(string.Format("PublicExponent=\t[{0}]\nPrivateExponent=[{1}]\nP=\t\t[{2}]\nQ=\t\t[{3}]",
           keyPair.PrivateKey.PublicExponent.ToString(),
           keyPair.PrivateKey.PrivateExponent.ToString(),
           keyPair.PrivateKey.P.ToString(),
           keyPair.PrivateKey.Q.ToString() ));
        Console.WriteLine();
        */
 
        byte[] signature = generateRSASignature(keyPair.PrivateKey,ExValues.sampleSignatureMessage);
        string  intSignature = new BigInteger(signature).ToString();       
        Console.WriteLine(string.Format("PKCS#1.5 signature:\t[{0}]",
            intSignature.Substring(1,20)+"..." + intSignature.Substring(intSignature.Length-20,20)));

        bool result = verifyRSAignature(keyPair.PublicKey, signature,ExValues.sampleSignatureMessage); 
        Console.WriteLine(string.Format("PKCS#1.5 verification:\tplainTextData=[{0}]\tVerification=[{1}]",
           System.Text.Encoding.Default.GetString(ExValues.sampleSignatureMessage),
           result ? "True" : "False"));


        Console.Write("\nSetting a default FipsSecureRandom object for X9.31 signatures ... ");
        DrbgExamples.setDefaultDrbg();
        Console.WriteLine("done\n");

        signature = generateX931Signature(keyPair.PrivateKey,ExValues.sampleSignatureMessage);
        intSignature = new BigInteger(signature).ToString();       
        Console.WriteLine(string.Format("X9.31 signature:\t[{0}]",
            intSignature.Substring(1,20)+"..." + intSignature.Substring(intSignature.Length-20,20)));
        result = verifyX931Signature(keyPair.PublicKey, signature,ExValues.sampleSignatureMessage);
        Console.WriteLine(string.Format("X9.31 verification:\tplainTextData=[{0}]\tVerification=[{1}]",
           System.Text.Encoding.Default.GetString(ExValues.sampleSignatureMessage),
           result ? "True" : "False"));
        Console.WriteLine();
        
        signature = generatePSSSignature(testPrivateKey,ExValues.sampleSignatureMessage);
        intSignature = new BigInteger(signature).ToString();       
        Console.WriteLine(string.Format("PSS signature:\t\t[{0}]",
            intSignature.Substring(1,20)+"..." + intSignature.Substring(intSignature.Length-20,20)));
        result = verifyPSSSignature(testPublicKey, signature,ExValues.sampleSignatureMessage);
        Console.WriteLine(string.Format("PSS verification:\tplainTextData=[{0}]\tVerification=[{1}]",
           System.Text.Encoding.Default.GetString(ExValues.sampleSignatureMessage),
           result ? "True" : "False"));
        Console.WriteLine();

        signature = generatePSSSignatureWithParameters(testPrivateKey,ExValues.sampleSignatureMessage);
        intSignature = new BigInteger(signature).ToString();       
        Console.WriteLine(string.Format("PSS signature:\t\t[{0}]",
            intSignature.Substring(1,20)+"..." + intSignature.Substring(intSignature.Length-20,20)));
        result = verifyPSSSignatureWithParameters(testPublicKey, signature,ExValues.sampleSignatureMessage);
        Console.WriteLine(string.Format("PSS verification:\tplainTextData=[{0}]\tVerification=[{1}]",
           System.Text.Encoding.Default.GetString(ExValues.sampleSignatureMessage),
           result ? "True" : "False"));
        Console.WriteLine();


        //Console.WriteLine("RSA OAEP key wrapping");
        //public static byte[] oaepKeyWrap(AsymmetricRsaPublicKey key, byte[] inputData)
        AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey>  keyPairWrapping = 
            generateRSAKeyPair(BigInteger.ValueOf(0x10001), 2048);

        /// <summary>
        /// Showing how to wrap key using RSA OAEP
        /// </summary>
        //byte[] cipherTextData = oaepKeyWrap(keyPairWrapping.PublicKey, longBytes);
        byte[] cipherTextData = oaepKeyWrap(keyPairWrapping.PublicKey, ExValues.sampleUnpaddedKey);
        Console.WriteLine(string.Format("RSA OAEP Key Wrapping   RSA Modulus key=[{0}]  inputData=[{1}]    wrappedData=[{2}]",
           keyPairWrapping.PublicKey.Modulus.ToString().Substring(1,20) + "..." + 
           keyPairWrapping.PublicKey.Modulus.ToString().Substring(keyPairWrapping.PrivateKey.Modulus.ToString().Length-20,20),
           Encoding.Default.GetString(ExValues.sampleUnpaddedKey),
           string.Join(",",cipherTextData).Substring(0,43) + "..."));
        byte[] plainTextData = oaepKeyUnwrap(keyPairWrapping.PrivateKey, cipherTextData);
        Console.WriteLine(string.Format("RSA OAEP Key Unwrapping RSA Modulus key=[{0}]  cipherData=[{1}]    unwrappedData=[{2}]",
           keyPairWrapping.PrivateKey.Modulus.ToString().Substring(1,20) + "..." + 
           keyPairWrapping.PrivateKey.Modulus.ToString().Substring(keyPairWrapping.PrivateKey.Modulus.ToString().Length-20,20),
           string.Join(",",cipherTextData).Substring(0,43) + "...",
           Encoding.Default.GetString(plainTextData)));


        /// <summary>
        /// Showing how to wrap key using RSA OAEP with parameters
        /// </summary>
        cipherTextData = oaepKeyWrapWithParameters(keyPairWrapping.PublicKey, ExValues.sampleUnpaddedKey);
        Console.WriteLine(string.Format("RSA OAEP Wrap With Params    RSA Modulus key=[{0}]  inputData=[{1}]    wrappedData=[{2}]",
           keyPairWrapping.PublicKey.Modulus.ToString().Substring(1,20) + "..." + 
           keyPairWrapping.PublicKey.Modulus.ToString().Substring(keyPairWrapping.PrivateKey.Modulus.ToString().Length-20,20),
           Encoding.Default.GetString(ExValues.sampleUnpaddedKey),
           string.Join(",",cipherTextData).Substring(0,43) + "..."));
        plainTextData = oaepKeyUnwrapWithParameters(keyPairWrapping.PrivateKey, cipherTextData);
        Console.WriteLine(string.Format("RSA OAEP Unwrap With Params  RSA Modulus key=[{0}]  cipherData=[{1}]    unwrappedData=[{2}]",
           keyPairWrapping.PrivateKey.Modulus.ToString().Substring(1,20) + "..." + 
           keyPairWrapping.PrivateKey.Modulus.ToString().Substring(keyPairWrapping.PrivateKey.Modulus.ToString().Length-20,20),
           string.Join(",",cipherTextData).Substring(0,43) + "...",
           Encoding.Default.GetString(plainTextData)));


        Console.WriteLine("=================================");
        Console.WriteLine("End   Of RsaExamples.Main");
        Console.WriteLine("=================================");
        Console.WriteLine();
        Console.WriteLine();        
    }

    /// <summary>
    /// Example 26 – RSA Key Pair Generation
    /// </summary>    
    public static AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey> 
        generateRSAKeyPair(BigInteger e, int keySize)  
    {   

        //Console.WriteLine(string.Format("P=[{0}]\tQ=[{1}]\tG=[{2}]",
        //   dsaDomainParams.P.ToString(),
        //   dsaDomainParams.Q.ToString(),
        //   dsaDomainParams.G.ToString()));


        FipsRsa.KeyGenerationParameters keyGenParameters =
            new FipsRsa.KeyGenerationParameters(e, keySize);
                
        FipsRsa.KeyPairGenerator kpGen = 
            CryptoServicesRegistrar.CreateGenerator(keyGenParameters, new SecureRandom());

        return kpGen.GenerateKeyPair();
    }  

    // This shows how you can use the For(...) method in the KeyGenerationParameters
    // class, to "label" your keys. Note that, there is no difference in the
    // actual algorithm for generating the RSA keys, it simply allows you to
    // determine the purpose of the keys if you need this.
    public static AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey>  
        generateX931KeyPair(BigInteger e, int keySize)  
    {   
        FipsRsa.KeyGenerationParameters keyGenParameters = 
            new FipsRsa.KeyGenerationParameters(e, keySize).For(FipsRsa.X931);
               
        FipsRsa.KeyPairGenerator kpGen = 
            CryptoServicesRegistrar.CreateGenerator(keyGenParameters, new SecureRandom());

        return kpGen.GenerateKeyPair();
    }  
    
    /// <summary>
    /// Example 27 – The PKCS#1.5 Signature 
    /// </summary>    
    public static byte[] generateRSASignature(AsymmetricRsaPrivateKey key, byte[] message)
    {
        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService(key, new SecureRandom());

        ISignatureFactory<FipsRsa.SignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(FipsRsa.Pkcs1v15);

        IStreamCalculator<IBlockResult> calculator = signer.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().Collect();
    }
       
    /// <summary>
    /// Example 27 – The PKCS#1.5 verification
    /// </summary>    
    public static bool verifyRSAignature(AsymmetricRsaPublicKey key, byte[] signature, byte[] message)
    {
        IVerifierFactoryService verifierFactoryProvider =
            CryptoServicesRegistrar.CreateService(key);

        IVerifierFactory<FipsRsa.SignatureParameters> verifier = 
            verifierFactoryProvider.CreateVerifierFactory(FipsRsa.Pkcs1v15);

        IStreamCalculator<IVerifier> calculator = verifier.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().IsVerified(signature);
    }       
    
    
    /// <summary>
    /// Example 28 – The X9.31 Signature 
    /// </summary>    
    public static byte[] generateX931Signature(AsymmetricRsaPrivateKey key, byte[] message)
    {
        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService(key);

        ISignatureFactory<FipsRsa.SignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(FipsRsa.X931);

        IStreamCalculator<IBlockResult> calculator = signer.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().Collect();
    }  

    /// <summary>
    /// Example 28 – The X9.31 verification
    /// </summary>    
    public static bool verifyX931Signature(AsymmetricRsaPublicKey key, byte[] signature, byte[] message)
    {
        IVerifierFactoryService verifierFactoryProvider =
            CryptoServicesRegistrar.CreateService(key);

        IVerifierFactory<FipsRsa.SignatureParameters> verifier = 
            verifierFactoryProvider.CreateVerifierFactory(FipsRsa.X931);

        IStreamCalculator<IVerifier> calculator = verifier.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().IsVerified(signature);
    }       
    
    /// <summary>
    /// Example 29 – The PSS Signature 
    /// </summary>    
    public static byte[] generatePSSSignature(AsymmetricRsaPrivateKey key, byte[] message)
    {
        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService(key);

        ISignatureFactory<FipsRsa.PssSignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(FipsRsa.Pss.WithSaltLength(20));

        IStreamCalculator<IBlockResult> calculator = signer.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().Collect();
    }    


    /// <summary>
    /// Example 29 – The PSS verification
    /// </summary>    
    public static bool verifyPSSSignature(AsymmetricRsaPublicKey key, byte[] signature, byte[] message)
    {
        IVerifierFactoryService verifierFactoryProvider =
            CryptoServicesRegistrar.CreateService(key);

        IVerifierFactory<FipsRsa.PssSignatureParameters> verifier = 
            verifierFactoryProvider.CreateVerifierFactory(FipsRsa.Pss.WithSaltLength(20));

        IStreamCalculator<IVerifier> calculator = verifier.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().IsVerified(signature);
    }       
        
        
    /// <summary>
    /// Example 30 – The PSS Signature with parameters
    /// </summary>    
    public static byte[] generatePSSSignatureWithParameters(AsymmetricRsaPrivateKey key, byte[] message)
    {
        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService(key);

        ISignatureFactory<FipsRsa.PssSignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(
                FipsRsa.Pss.WithDigest(FipsShs.Sha384).WithMgfDigest(
                    FipsShs.Sha384).WithSalt(ExValues.sampleInput));

        IStreamCalculator<IBlockResult> calculator = signer.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().Collect();
    }    


    /// <summary>
    /// Example 30 – The PSS verification with Parameters
    /// </summary>    
    public static bool verifyPSSSignatureWithParameters(AsymmetricRsaPublicKey key, 
        byte[] signature, byte[] message)
    {
        IVerifierFactoryService verifierFactoryProvider =
            CryptoServicesRegistrar.CreateService(key);

        IVerifierFactory<FipsRsa.PssSignatureParameters> verifier = 
            verifierFactoryProvider.CreateVerifierFactory(
                FipsRsa.Pss.WithDigest(FipsShs.Sha384).WithMgfDigest(
                    FipsShs.Sha384).WithSalt(ExValues.sampleInput));

        IStreamCalculator<IVerifier> calculator = verifier.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().IsVerified(signature);
    }       
        

    /// <summary>
    /// Example 36 – wrapping and unwrapping using OAEP RSA
    /// </summary>    
    public static byte[] oaepKeyWrap(AsymmetricRsaPublicKey key, byte[] inputData)
    {
        IKeyWrappingService provider = CryptoServicesRegistrar.CreateService(key, new SecureRandom());
        IKeyWrapper<FipsRsa.OaepWrapParameters> wrapper = provider.CreateKeyWrapper(FipsRsa.WrapOaep);
        byte[] wrappedInput = wrapper.Wrap(inputData).Collect();
        return wrappedInput;
    }
    
    public static byte[]  oaepKeyUnwrap(AsymmetricRsaPrivateKey key, byte[] cipherToUnwrap)
    {
        IKeyUnwrappingService provider = CryptoServicesRegistrar.CreateService(key);

        IKeyUnwrapper<FipsRsa.OaepWrapParameters> unwrapper = provider.CreateKeyUnwrapper(FipsRsa.WrapOaep);
        byte[] inputData = unwrapper.Unwrap(cipherToUnwrap, 0, cipherToUnwrap.Length).Collect();
        return inputData;
    }
    
    /// <summary>
    /// Example 37 – wrapping and unwrapping using OAEP RSA with parameters
    /// </summary>    
    public static byte[] oaepKeyWrapWithParameters(AsymmetricRsaPublicKey key, byte[] inputData)
    {
        IKeyWrappingService provider = CryptoServicesRegistrar.CreateService(key, new SecureRandom());
        IKeyWrapper<FipsRsa.OaepWrapParameters> wrapper = 
            provider.CreateKeyWrapper(
                FipsRsa.WrapOaep.WithDigest(
                    FipsShs.Sha512).WithMgfDigest(
                        FipsShs.Sha512).WithEncodingParams(ExValues.sampleNonce));        
        byte[] wrappedInput = wrapper.Wrap(inputData).Collect();
        return wrappedInput;
    }
    
    public static byte[]  oaepKeyUnwrapWithParameters(AsymmetricRsaPrivateKey key, byte[] cipherToUnwrap)
    {
        IKeyUnwrappingService provider = CryptoServicesRegistrar.CreateService(key);

        IKeyUnwrapper<FipsRsa.OaepWrapParameters> unwrapper = 
            provider.CreateKeyUnwrapper(
                FipsRsa.WrapOaep.WithDigest(
                    FipsShs.Sha512).WithMgfDigest(
                        FipsShs.Sha512).WithEncodingParams(ExValues.sampleNonce));        
        byte[] inputData = unwrapper.Unwrap(cipherToUnwrap, 0, cipherToUnwrap.Length).Collect();
        return inputData;
    }
    
    
}
