using System;
using System.Text;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;

using bcfipsin100.util;

public class ShsExamples
{
    
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("=================================");
        Console.WriteLine("Start Of ShsExamples.Main (HASHs)");
        Console.WriteLine("=================================");

        byte[] result;
        
        result = sha512Digest(Encoding.Default.GetBytes("Once Upon a midnight dreary"));
        Console.WriteLine(string.Format("SHA512\t\tplainTextData=[{0}]\tHashTextData=[{1}]",
           "Once Upon a midnight dreary",
           Hex.ToHexString(result).Substring(0,80) + "..." + Hex.ToHexString(result).Substring(120,8)));
       
        result = sha3_512Digest(Encoding.Default.GetBytes("SHA3_512 TEST"));
        Console.WriteLine(string.Format("SHA3_512_256\tplainTextData=[{0}]\tHashTextData=[{1}]",
           "Once Upon a midnight dreary",
           Hex.ToHexString(result).Substring(0,80) + "..." + Hex.ToHexString(result).Substring(120,8)));
        Console.WriteLine();

        result = shake256Output(Encoding.Default.GetBytes("Once Upon a midnight dreary"));
        Console.WriteLine(string.Format("SHAKE256\tplainTextData=[{0}]\tHashTextData=[{1}]",
           "Once Upon a midnight dreary",
           Hex.ToHexString(result)));
           
        result = shake256OutputContinuous(Encoding.Default.GetBytes("Once Upon a midnight dreary"));
        Console.WriteLine(string.Format("SHAKE256\tplainTextData=[{0}]\tHashTextData=[{1}]",
           "Once Upon a midnight dreary",
           Hex.ToHexString(result)));
        Console.WriteLine("CONTINUOUS OUTPUT");
        Console.WriteLine();
        
        byte[] hmacKey = generateHmacKey();       
        Console.WriteLine(string.Format("HMAC KEY GENERATION\t\t\t\t\t\tHmacKey=[{0}]",
           Hex.ToHexString(hmacKey)));
        Console.WriteLine();

        result = calculateHMAC(Encoding.Default.GetBytes("Once Upon a midnight dreary"),hmacKey);
        Console.WriteLine(string.Format("HMAC CALC\tplainTextData=[{0}]\tHashTextData=[{1}]",
           "Once Upon a midnight dreary",
           Hex.ToHexString(result).Substring(0,80) + "..." + Hex.ToHexString(result).Substring(120,8)));
        Console.WriteLine();
           

        //Console.WriteLine(string.Format("HMAC SYM KEY rawKey=[{0}]\t hmacKey=[{1}]",
        //   "000102030405060708090a0b0c0d0e0f10111213",
        //   Hex.ToHexString(ExValues.sampleHMacKey.GetKeyBytes())));

        result = calculateCMAC(Encoding.Default.GetBytes("Once Upon a midnight dreary"),ExValues.sampleAesKeyInput);
        Console.WriteLine(string.Format("CMAC CALC\tplainTextData=[{0}]\tHashTextData=[{1}]",
           "Once Upon a midnight dreary",
           Hex.ToHexString(result)));
           
        result = calculateGMAC(Encoding.Default.GetBytes("Once Upon a midnight dreary"),ExValues.sampleAesKeyInput);
        Console.WriteLine(string.Format("GMAC CALC\tplainTextData=[{0}]\tHashTextData=[{1}]",
           "Once Upon a midnight dreary",
           Hex.ToHexString(result)));


        result = calculateCBCMAC(Encoding.Default.GetBytes("Once Upon a midnight dreary"),ExValues.sampleAesKeyInput);
        Console.WriteLine(string.Format("CBCMAC CALC\tplainTextData=[{0}]\tHashTextData=[{1}]",
           "Once Upon a midnight dreary",
           Hex.ToHexString(result)));
        Console.WriteLine();

        result = gcmEncrypt(Encoding.Default.GetBytes("Once Upon a midnight dreary"),ExValues.sampleAesKeyInput);
        Console.WriteLine(string.Format("GCM CALC\tplainTextData=[{0}]\tHashTextData=[{1}]",
           "Once Upon a midnight dreary",
           Hex.ToHexString(result)));
        Console.WriteLine("GCM CIPHER WITH ONLY AAD NO PLAIN TEXT DATA -- SAME AS GMAC CALC");

        Console.WriteLine("=================================");
        Console.WriteLine("End   Of ShsExamples.Main (HASHs)");
        Console.WriteLine("=================================");
        Console.WriteLine();
        Console.WriteLine();        
    }
    

    
    /// <summary>
    /// Example 14 – SHA512 Digest Example
    /// </summary>
    public static byte[] sha512Digest(byte[] message)  
    {   
        IDigestFactory<FipsShs.Parameters> provider = CryptoServicesRegistrar.CreateService(FipsShs.Sha512); 
        IStreamCalculator<IBlockResult> calculator = provider.CreateCalculator();
        
        Stream messageStream = calculator.Stream;
        messageStream.Write(message, 0, message.Length);
        messageStream.Close();
        return calculator.GetResult().Collect();
    } 
    
        
    /// <summary>
    /// Example 14 – SHA3 512 Digest Example
    /// </summary>
    public static byte[] sha3_512Digest(byte[] message)  
    {   
        IDigestFactory<FipsShs.Parameters> provider = CryptoServicesRegistrar.CreateService(FipsShs.Sha3_512); 
        IStreamCalculator<IBlockResult> calculator = provider.CreateCalculator();       
        Stream messageStream = calculator.Stream;
        messageStream.Write(message, 0, message.Length);
        messageStream.Close();
        return calculator.GetResult().Collect();
    } 

    /// <summary>
    /// Example 15 – Basic Use of an XOF
    /// </summary>
    public static byte[] shake256Output(byte[] message, int outputLength = 32)  
    {
        IXofFactory<FipsShs.XofParameters> provider =  CryptoServicesRegistrar.CreateService(FipsShs.Shake256);
        IVariableStreamCalculator<IBlockResult> calculator = provider.CreateCalculator();
        Stream messageStream = calculator.Stream;
        messageStream.Write(message, 0, message.Length);
        messageStream.Close();
        return calculator.GetResult(outputLength).Collect();
    }
    

    /// <summary>
    /// Example 16 – Continuous use of an XOF
    /// </summary>
    public static byte[] shake256OutputContinuous(byte[] message, int outputLength = 16)  
    {
        IXofFactory<FipsShs.XofParameters> provider =  CryptoServicesRegistrar.CreateService(FipsShs.Shake256);
        IVariableStreamCalculator<IBlockResult> calculator = provider.CreateCalculator();
        Stream messageStream = calculator.Stream;
        messageStream.Write(message, 0, message.Length);
        messageStream.Close();
        return calculator.GetResult(outputLength).Collect().Concat(
                   calculator.GetResult(outputLength).Collect()).ToArray();
    }


    /// <summary>
    /// Example 17 – SecretKey generation for HMAC
    /// </summary>
    
    public static byte[] generateHmacKey()
    {
        FipsSecureRandom  hmacSecureRandom = DrbgExamples.buildDrbgForKeys();
        byte[] hmacKey = new byte[32];
        hmacSecureRandom.NextBytes(hmacKey);
        return hmacKey;
    }

    /// <summary>
    /// Example 18 – HMAC calculation
    /// </summary>
    public static byte[] calculateHMAC(byte[] message, byte[] hmacKey)
    {
        IMacFactoryService  hmacProvider =  CryptoServicesRegistrar.CreateService(
            new FipsShs.Key(FipsShs.Sha512HMac, hmacKey));
        IMacFactory<FipsShs.AuthenticationParameters>  hmacFactory = hmacProvider.CreateMacFactory(FipsShs.Sha512HMac);
        IStreamCalculator<IBlockResult> calculator = hmacFactory.CreateCalculator();
        Stream messageStream = calculator.Stream;
        messageStream.Write(message, 0, message.Length);
        messageStream.Close();
        return calculator.GetResult().Collect();        
    }

    /// <summary>
    /// Example 19 – MAC Calculation using CMAC
    /// </summary>    
    public static byte[] calculateCMAC(byte[] message, byte[] aesKeyInput)
    {
        IMacFactoryService  cmacProvider =  CryptoServicesRegistrar.CreateService(
            new FipsAes.Key(FipsAes.CMac, aesKeyInput));
            
        IMacFactory<FipsAes.AuthenticationParameters>  cmacFactory = 
            cmacProvider.CreateMacFactory(FipsAes.CMac.WithMacSize(128));
            
        IStreamCalculator<IBlockResult> calculator = cmacFactory.CreateCalculator();
        Stream messageStream = calculator.Stream;
        messageStream.Write(message, 0, message.Length);
        messageStream.Close();
        return calculator.GetResult().Collect();        
    }


    /// <summary>
    /// Example 20 – MAC Calculation using GMAC
    /// </summary>    
    public static byte[] calculateGMAC(byte[] message, byte[] aesKeyInput)
    {        

        IMacFactoryService  gmacProvider =  CryptoServicesRegistrar.CreateService(
            new FipsAes.Key(FipsAes.GMac, aesKeyInput));

        IMacFactory<FipsAes.AuthenticationParametersWithIV>  gmacFactory = 
            gmacProvider.CreateMacFactory(FipsAes.GMac.WithIV(ExValues.sampleIVcounter).WithMacSize(128));

        IStreamCalculator<IBlockResult> calculator = gmacFactory.CreateCalculator();
        Stream messageStream = calculator.Stream;
        messageStream.Write(message, 0, message.Length);
        messageStream.Close();
        return calculator.GetResult().Collect();        
    }        

    /// <summary>
    /// Example 21 – MAC Calculation using CCM        
    /// </summary>    
    public static byte[] calculateCBCMAC(byte[] message, byte[] aesKeyInput)
    {        

        IMacFactoryService  ccmProvider =  CryptoServicesRegistrar.CreateService(
            new FipsAes.Key(FipsAes.Ccm, aesKeyInput));

        IMacFactory<FipsAes.AuthenticationParametersWithIV>  ccmFactory = 
            ccmProvider.CreateMacFactory(FipsAes.Ccm.WithIV(ExValues.sampleIVcounter).WithMacSize(128));

        IStreamCalculator<IBlockResult> calculator = ccmFactory.CreateCalculator();
        Stream messageStream = calculator.Stream;
        messageStream.Write(message, 0, message.Length);
        messageStream.Close();
        return calculator.GetResult().Collect();        
    }         


    /// <summary>
    /// Example 20 – MAC Calculation using GCM WITH NO PLAIN TEXT DATA
    /// </summary>    
    public static byte[] gcmEncrypt(byte[] message, byte[] aesKeyInput)
    {        
        IAeadCipherService provider = CryptoServicesRegistrar.CreateService(
            new FipsAes.Key(FipsAes.Gcm, aesKeyInput)); 
        IAeadCipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateAeadEncryptorBuilder(FipsAes.Gcm.WithIV(ExValues.sampleIVcounter));

        MemoryOutputStream streamOutput = new MemoryOutputStream();
        
        
        IAeadCipher noPaddedEncryptor = encryptBuilder.BuildAeadCipher(AeadUsage.AAD_FIRST, streamOutput);
        noPaddedEncryptor.AadStream.Write(message,0,message.Length);
               
        noPaddedEncryptor.Stream.Close();

        byte[] cipherMacData = noPaddedEncryptor.GetMac().Collect();
        
        return cipherMacData;
    }
}
