using System;
using System.Text;
using System.Runtime.InteropServices;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Asymmetric;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;
using Org.BouncyCastle.Operators;
using Org.BouncyCastle.Asn1;
using Org.BouncyCastle.Asn1.Crmf;
using Org.BouncyCastle.Asn1.Pkcs;
using Org.BouncyCastle.Asn1.Nist;
using Org.BouncyCastle.Asn1.X500;
using Org.BouncyCastle.Asn1.X509;
using Org.BouncyCastle.Cert;
using Org.BouncyCastle.Crmf;
using Org.BouncyCastle.Pkcs;
using Org.BouncyCastle.OpenSsl;
using Org.BouncyCastle.Ocsp;
using Org.BouncyCastle.Pkix;
using Org.BouncyCastle.Utilities.Date;
using bcfipsin100.util;

public class CertExamples
{
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("===============================");
        Console.WriteLine("Start Certificate Examples.Main");
        Console.WriteLine("===============================");
 
        {
            //Example 51 – Building a Version 1 X.509 Certificate
            Console.WriteLine();
            Console.WriteLine("*** Example 51 – Building a Version 1 X.509 Certificate ***");
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairCA = 
                generateECKeyPair();
                
            X509Certificate x509CertV1 = makeX509V1Certificate(keyPairCA.PrivateKey, keyPairCA.PublicKey);
            StringWriter strWriterCA = new StringWriter();
            OpenSslPemWriter pemWriterCA = new OpenSslPemWriter(strWriterCA);
            pemWriterCA.WriteObject(x509CertV1);
            File.WriteAllText(@"testCA.pem",strWriterCA.ToString());

            {
                X500Name name1 = x509CertV1.SubjectDN;
                X500Name issuer1 = x509CertV1.IssuerDN;
                TbsCertificateStructure tbsCert = TbsCertificateStructure.GetInstance(
                        Asn1Object.FromByteArray(x509CertV1.GetTbsCertificate()));
                X500Name name2 = tbsCert.Subject;            
                X500Name issuer2 = tbsCert.Issuer;            
            }    
        
        
        
            //Example 52 – Building a Version 3 X.509 Certificate
            Console.WriteLine();
            Console.WriteLine("*** Example 52 – Building a Version 3 X.509 Certificate ***");
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairEE = 
                generateECKeyPair();
                
            X509Certificate x509CertEE = makeX509V3Certificate(x509CertV1,
                    keyPairCA.PrivateKey, keyPairEE.PublicKey);
            StringWriter strWriterEE = new StringWriter();
            OpenSslPemWriter pemWriterEE = new OpenSslPemWriter(strWriterEE);
            pemWriterEE.WriteObject(x509CertEE);
            File.WriteAllText(@"testEE.pem",strWriterEE.ToString());
        }
        
        {            
            //Example 53 – Creating a CRL 
            Console.WriteLine();
            Console.WriteLine("*** Example 53 – Creating a CRL ***");
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairCA = generateECKeyPair();                
            X509Certificate x509CertCA = makeX509V1Certificate(keyPairCA.PrivateKey, keyPairCA.PublicKey);
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairC1 = generateECKeyPair();                
            X509Certificate x509CertC1 = makeX509V3Certificate(x509CertCA, 
                        keyPairCA.PrivateKey, keyPairC1.PublicKey);
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairC2 = generateECKeyPair();                
            X509Certificate x509CertC2 = makeX509V3Certificate(x509CertCA, 
                        keyPairCA.PrivateKey, keyPairC2.PublicKey);
                        
            X509Certificate[] revoke = new X509Certificate[2]{x509CertC1, x509CertC2};
            X509Crl crl = CreateCRL(x509CertCA,keyPairCA.PrivateKey,revoke);
            StringWriter strWriter = new StringWriter();
            OpenSslPemWriter pemWriter = new OpenSslPemWriter(strWriter);
            pemWriter.WriteObject(crl);
            File.WriteAllText(@"testcrl.pem",strWriter.ToString());            
        }
                
        {            
            //Example 54 – Creating an OCSP Request 
            Console.WriteLine();
            Console.WriteLine("*** Example 54 – Creating an OCSP Request ***");
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairCA = generateECKeyPair();                
            X509Certificate x509CertCA = makeX509V1Certificate(keyPairCA.PrivateKey, keyPairCA.PublicKey);
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairC1 = generateECKeyPair();                
            X509Certificate x509CertC1 = makeX509V3Certificate(x509CertCA, 
                        keyPairCA.PrivateKey, keyPairC1.PublicKey);
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairC2 = generateECKeyPair();                
            X509Certificate x509CertC2 = makeX509V3Certificate(x509CertCA, 
                        keyPairCA.PrivateKey, keyPairC2.PublicKey);
            
            X509Certificate[] certsToCheck = new X509Certificate[2]{x509CertC1, x509CertC2};
            OcspReq request = makeOcspRequest(x509CertCA, certsToCheck);
        }
        
        {
            //Example 55 – Creating an OCSP Response
            Console.WriteLine();
            Console.WriteLine("*** Example 55 – Creating an OCSP Response ***");

            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairCA = generateECKeyPair();                
            X509Certificate x509CertCA = makeX509V1Certificate(keyPairCA.PrivateKey, keyPairCA.PublicKey);
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairC1 = generateECKeyPair();                
            X509Certificate x509CertC1 = makeX509V3Certificate(x509CertCA, 
                        keyPairCA.PrivateKey, keyPairC1.PublicKey);
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairC2 = generateECKeyPair();                
            X509Certificate x509CertC2 = makeX509V3Certificate(x509CertCA, 
                        keyPairCA.PrivateKey, keyPairC2.PublicKey);
            
            X509Certificate[] certsToCheck = new X509Certificate[2]{x509CertC1, x509CertC2};
            OcspReq request = makeOcspRequest(x509CertCA, certsToCheck);

            BasicOcspResp response = makeOcspResponse(request, x509CertCA, keyPairCA.PrivateKey);
            
        }
        
        {
            //Example 56 – Checking an OCSP Response
            Console.WriteLine();
            Console.WriteLine("*** Example 56 – Checking an OCSP Response ***");
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairCA = generateECKeyPair();                
            X509Certificate x509CertCA = makeX509V1Certificate(keyPairCA.PrivateKey, keyPairCA.PublicKey);
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairC1 = generateECKeyPair();                
            X509Certificate x509CertC1 = makeX509V3Certificate(x509CertCA, 
                        keyPairCA.PrivateKey, keyPairC1.PublicKey);
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairC2 = generateECKeyPair();                
            X509Certificate x509CertC2 = makeX509V3Certificate(x509CertCA, 
                        keyPairCA.PrivateKey, keyPairC2.PublicKey);
            
            X509Certificate[] certsToCheck = new X509Certificate[2]{x509CertC1, x509CertC2};
            OcspReq request = makeOcspRequest(x509CertCA, certsToCheck);

            BasicOcspResp response = makeOcspResponse(request, x509CertCA, keyPairCA.PrivateKey);
            
            bool[] result = checkCertificates(response, x509CertCA, certsToCheck);
            Console.WriteLine(string.Format("result:[{0}]",string.Join(",",result)));            
        }

        {
            //Example 57 – Basic CertPath Validation
            Console.WriteLine();
            Console.WriteLine("*** Example 57 – Basic CertPath Validation ***");
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairTrustAnchor = generateECKeyPair();                
            X509Certificate trustAnchor = makeX509V3CertificateTrustAnchor(keyPairTrustAnchor.PrivateKey, keyPairTrustAnchor.PublicKey);
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairCA = generateECKeyPair();                
            X509Certificate x509CertCA = makeX509V3Certificate(trustAnchor, keyPairTrustAnchor.PrivateKey, keyPairCA.PublicKey);
            
            AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairEE = generateECKeyPair();                
            X509Certificate x509CertEE = makeX509V3Certificate(x509CertCA, keyPairCA.PrivateKey, keyPairEE.PublicKey);
            
            PkixCertPathValidatorResult result = validateCertPath(trustAnchor, x509CertCA, x509CertEE);
        }     
        
        Console.WriteLine("===============================");
        Console.WriteLine("End   Certificate Examples.Main");
        Console.WriteLine("===============================");
        Console.WriteLine();
        Console.WriteLine();        
    }

    public static AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  generateECKeyPair()
    {
        FipsEC.KeyGenerationParameters keyGenParametersCA =
            new FipsEC.KeyGenerationParameters(FipsEC.DomainParams.K409).For(FipsEC.Cdh);
        ECDomainParameters ecParamsCA = keyGenParametersCA.DomainParameters;    
        FipsEC.KeyPairGenerator kpGenCA = 
            CryptoServicesRegistrar.CreateGenerator(keyGenParametersCA, ExValues.DHFixedRandomA(1000));
        return kpGenCA.GenerateKeyPair();
    }


    public static X509Certificate makeX509V1Certificate(
        IAsymmetricPrivateKey caSignerKey, IAsymmetricPublicKey caPublicKey)
    {
        X509V1CertificateGenerator certGen = new X509V1CertificateGenerator();
        BigInteger serialNumber = ExValues.FixedRandomBigInteger(32,1);
        certGen.SetSerialNumber(serialNumber);
        certGen.SetIssuerDN(new X500Name("CN=Issuer Self_Signed_CA"));
        certGen.SetNotBefore(DateTime.UtcNow.AddSeconds(-50));
        certGen.SetNotAfter(DateTime.UtcNow.AddSeconds(50));    
        certGen.SetSubjectDN(new X500Name("CN=Subject Self_Signed_CA"));
        certGen.SetPublicKey((AsymmetricECPublicKey)caPublicKey);

        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService((AsymmetricECPrivateKey)caSignerKey, 
                ExValues.DHFixedRandomA(1000));

        ISignatureFactory<FipsEC.SignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(FipsEC.Dsa);
            
        return certGen.Generate(new PkixSignatureFactory(signer));
            
    }
    
    public static X509Certificate makeX509V3CertificateTrustAnchor(
        IAsymmetricPrivateKey caSignerKey, IAsymmetricPublicKey caPublicKey)
    {
        X509V3CertificateGenerator certGen = new X509V3CertificateGenerator();
        BigInteger serialNumber = ExValues.FixedRandomBigInteger(32,1);
        certGen.SetSerialNumber(serialNumber);
        certGen.SetIssuerDN(new X500Name("CN=Issuer Self_Signed_CA"));
        certGen.SetNotBefore(DateTime.UtcNow.AddSeconds(-50));
        certGen.SetNotAfter(DateTime.UtcNow.AddSeconds(50));    
        certGen.SetSubjectDN(new X500Name("CN=Subject Self_Signed_CA"));
        certGen.SetPublicKey((AsymmetricECPublicKey)caPublicKey);

        certGen.AddExtension(X509Extensions.SubjectKeyIdentifier, false, new SubjectKeyIdentifier(caPublicKey.GetEncoded()));
        certGen.AddExtension(X509Extensions.BasicConstraints, true, new BasicConstraints(true));
        BasicConstraints bc = new BasicConstraints(true);
        certGen.AddExtension(X509Extensions.KeyUsage, true, new KeyUsage(KeyUsage.DigitalSignature | KeyUsage.KeyEncipherment));

        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService((AsymmetricECPrivateKey)caSignerKey, 
                ExValues.DHFixedRandomA(1000));

        ISignatureFactory<FipsEC.SignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(FipsEC.Dsa);
            
        return certGen.Generate(new PkixSignatureFactory(signer));
            
    }

    
    public static X509Certificate makeX509V3Certificate(
        X509Certificate caCert,
        IAsymmetricPrivateKey caSignerKey, 
        IAsymmetricPublicKey endEntityPublicKey)
    {
        X509V3CertificateGenerator certGen = new X509V3CertificateGenerator();
        BigInteger serialNumber = ExValues.FixedRandomBigInteger(32,1);
        certGen.SetSerialNumber(serialNumber);
        certGen.SetIssuerDN(caCert.SubjectDN);
        certGen.SetNotBefore(DateTime.UtcNow.AddSeconds(-50));
        certGen.SetNotAfter(DateTime.UtcNow.AddSeconds(50));    
        certGen.SetSubjectDN(new X500Name("CN=End Entity Certifcate"));
        certGen.SetPublicKey((AsymmetricECPublicKey)endEntityPublicKey);

        certGen.AddExtension(X509Extensions.AuthorityKeyIdentifier, false, new AuthorityKeyIdentifier(caCert.GetEncoded()));
        certGen.AddExtension(X509Extensions.SubjectKeyIdentifier, false, new SubjectKeyIdentifier(endEntityPublicKey.GetEncoded()));
        certGen.AddExtension(X509Extensions.BasicConstraints, true, new BasicConstraints(true));

        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService((AsymmetricECPrivateKey)caSignerKey, 
                ExValues.DHFixedRandomA(1000));

        ISignatureFactory<FipsEC.SignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(FipsEC.Dsa);
            
        return certGen.Generate(new PkixSignatureFactory(signer));
        
    }    
    
    
    public static X509Crl CreateCRL(X509Certificate caCert,
            IAsymmetricPrivateKey caSignerKey, X509Certificate[] revokedCerts)
    {
        X509V2CrlGenerator crlGen = new X509V2CrlGenerator();
        CrlNumber crlNumber = new CrlNumber(ExValues.FixedRandomBigInteger(32,1));
        
        crlGen.SetIssuerDN(caCert.SubjectDN);
        crlGen.SetThisUpdate(DateTime.UtcNow);
        crlGen.SetNextUpdate(DateTime.UtcNow.AddSeconds(100));

        crlGen.AddExtension(X509Extensions.AuthorityKeyIdentifier, true, 
            new AuthorityKeyIdentifier(caCert.GetEncoded()));
        crlGen.AddExtension(X509Extensions.CrlNumber, false, crlNumber);
        

        for (int j = 0; j < revokedCerts.Length; ++j) 
        {
            crlGen.AddCrlEntry(revokedCerts[j].SerialNumber, 
                DateTime.UtcNow, CrlReason.PrivilegeWithdrawn);
        }

        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService((AsymmetricECPrivateKey)caSignerKey, 
                ExValues.DHFixedRandomA(1000));

        ISignatureFactory<FipsEC.SignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(FipsEC.Dsa);

        return crlGen.Generate(new PkixSignatureFactory(signer));

    }
        
        
    public static OcspReq makeOcspRequest(X509Certificate caCert, X509Certificate[] certsToCheck)
    {
        IDigestFactory<AlgorithmIdentifier> sha1Factory = 
            new PkixDigestFactoryProvider().CreateDigestFactory(CertificateID.HashSha1);
            
        OcspReqGenerator ocspGen = new OcspReqGenerator();

        for (int j = 0; j < certsToCheck.Length; ++j) 
        {
            CertificateID certID  = 
                new CertificateID(sha1Factory, caCert, certsToCheck[j].SerialNumber);        
            ocspGen.AddRequest(certID);
            Console.WriteLine(string.Format("CertID:[{0}]",certID.SerialNumber));            
        }
        return ocspGen.Generate();
    }        
    
    public static BasicOcspResp makeOcspResponse(OcspReq request, X509Certificate caCert, 
        IAsymmetricPrivateKey caSignerKey)
    {
		BasicOcspRespGenerator respGen = new BasicOcspRespGenerator(caCert.GetPublicKey());
        
        for (int j = 0; j < request.GetRequestList().Length; ++j) 
        {
            CertificateID certID = request.GetRequestList()[j].GetCertID();
            // ...
            // magic happens            
            // ...
            respGen.AddResponse(certID, CertificateStatus.Good);            
            Console.WriteLine(string.Format("CertID:[{0}]",certID.SerialNumber));            
        }

        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService((AsymmetricECPrivateKey)caSignerKey, 
                ExValues.DHFixedRandomA(1000));

        ISignatureFactory<FipsEC.SignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(FipsEC.Dsa);
	
        X509Certificate[] caCertChain = new X509Certificate[1]{caCert};
		BasicOcspResp response = respGen.Generate(
            new PkixSignatureFactory(signer), caCertChain, DateTime.UtcNow);
        return response;
    }

    public static bool[] checkCertificates(BasicOcspResp response, X509Certificate caCert,
                                           X509Certificate[] certsToCheck) 
    {

        bool[] result = new bool[certsToCheck.Length];
        for (int j = 0; j < response.Responses.Length; ++j) 
        { 
            result[j] = false; 
        }
        
        IVerifierFactoryProvider<AlgorithmIdentifier> verifierFactoryProvider = 
            new PkixVerifierFactoryProvider(caCert.GetPublicKey());

        Console.Write("Checking that the signature is valid ... ");            
        // Check that the signature is valid
        if (
            (!response.IsVerified(verifierFactoryProvider)) ||
            (response.Responses.Length != certsToCheck.Length)
           )
        {
            return result;
        }
        Console.WriteLine("done");            

        IDigestFactory<AlgorithmIdentifier> sha1Factory = 
            new PkixDigestFactoryProvider().CreateDigestFactory(CertificateID.HashSha1);
            
        for (int j = 0; j < response.Responses.Length; ++j) 
        {
            if (
                (response.Responses[j].GetCertStatus() == null) &&
                (response.Responses[j].GetCertID().SerialNumber.Equals(certsToCheck[j].SerialNumber)) &&
                (response.Responses[j].GetCertID().MatchesIssuer(sha1Factory,caCert))
               ) 
            {
                        Console.WriteLine("3000");
                        result[j] = true;
            }
        }
        
        return result;
    }        
    
    public static PkixCertPathValidatorResult validateCertPath(X509Certificate taCert, 
                        X509Certificate caCert, X509Certificate eeCert)
    {
        List<X509Certificate> certChain = new List<X509Certificate>();
        certChain.Add(eeCert);
        certChain.Add(caCert);
        PkixCertPath certPath = new PkixCertPath(certChain);
        
        HashSet<TrustAnchor> trustAnchors = new HashSet<TrustAnchor>();
        trustAnchors.Add(new TrustAnchor(taCert, null));
        
        PkixParameters param = new PkixParameters(trustAnchors);        
        param.Date = new DateTimeObject(DateTime.UtcNow);
        param.IsRevocationEnabled = false;
        PkixCertPathValidator certPathValidator = new PkixCertPathValidator();
        PkixCertPathValidatorResult result = certPathValidator.Validate(certPath, param);        
        return result;
    }
    
    public static PkixCertPathValidatorResult validateCertPathWithCrl(
                                   X509Certificate taCert, X509Crl taCrl, 
                                   X509Certificate caCert, X509Crl caCrl, 
                                   X509Certificate eeCert)
    {
        List<X509Certificate> certChain = new List<X509Certificate>();
        
        certChain.Add(eeCert);
        certChain.Add(caCert);
        PkixCertPath certPath = new PkixCertPath(certChain);
        
        HashSet<TrustAnchor> trustAnchors = new HashSet<TrustAnchor>();
        trustAnchors.Add(new TrustAnchor(taCert, null));

        List<X509Crl> x509Crls = new List<X509Crl>();
        x509Crls.Add(caCrl);
        x509Crls.Add(taCrl);
        X509CrlCollectionStore x509CrlStore = new X509CrlCollectionStore(x509Crls);

        PkixParameters param = new PkixParameters(trustAnchors);        
        param.Date = new DateTimeObject(DateTime.UtcNow);
        param.IsRevocationEnabled = true;
        param.AddCrlStore(x509CrlStore);
        PkixCertPathValidator certPathValidator = new PkixCertPathValidator();
        PkixCertPathValidatorResult result = certPathValidator.Validate(certPath, param);        
        return result;
    }    
    
    
}
