using System;
using System.Text;

using Org.BouncyCastle.Math;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Asymmetric;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;
using Org.BouncyCastle.Operators;
using Org.BouncyCastle.Asn1.Crmf;
using Org.BouncyCastle.Asn1.Pkcs;
using Org.BouncyCastle.Asn1.Nist;
using Org.BouncyCastle.Asn1.X500;
using Org.BouncyCastle.Asn1.X509;
using Org.BouncyCastle.Cert;
using Org.BouncyCastle.Crmf;
using Org.BouncyCastle.Pkcs;
using Org.BouncyCastle.OpenSsl;


using bcfipsin100.util;

public class CrmfExamples
{
 
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("============================");
        Console.WriteLine("Start Of   CrmfExamples.Main");
        Console.WriteLine("============================");

        AsymmetricRsaPrivateKey testPrivateKey = new AsymmetricRsaPrivateKey(FipsRsa.Alg,
                new BigInteger("495370a1fb18543c16d3631e3163255df62be6eee890d5f25509e4f778a8ea6fbbbcdf85dff64e0d972003ab3681fbba6dd41fd541829b2e582de9f2a4a4e0a2d0900bef4753db3cee0ee06c7dfae8b1d53b5953218f9cceea695b08668edeaadced9463b1d790d5ebf27e9115b46cad4d9a2b8efab0561b0810344739ada0733f", 16),
                new BigInteger("010001", 16),
                new BigInteger("6c66ffe98980c38fcdeab5159898836165f4b4b817c4f6a8d486ee4ea9130fe9b9092bd136d184f95f504a607eac565846d2fdd6597a8967c7396ef95a6eeebb4578a643966dca4d8ee3de842de63279c618159c1ab54a89437b6a6120e4930afb52a4ba6ced8a4947ac64b30a3497cbe701c2d6266d517219ad0ec6d347dbe9", 16),
                new BigInteger("08dad7f11363faa623d5d6d5e8a319328d82190d7127d2846c439b0ab72619b0a43a95320e4ec34fc3a9cea876422305bd76c5ba7be9e2f410c8060645a1d29edb", 16),
                new BigInteger("0847e732376fc7900f898ea82eb2b0fc418565fdae62f7d9ec4ce2217b97990dd272db157f99f63c0dcbb9fbacdbd4c4dadb6df67756358ca4174825b48f49706d", 16),
                new BigInteger("05c2a83c124b3621a2aa57ea2c3efe035eff4560f33ddebb7adab81fce69a0c8c2edc16520dda83d59a23be867963ac65f2cc710bbcfb96ee103deb771d105fd85", 16),
                new BigInteger("04cae8aa0d9faa165c87b682ec140b8ed3b50b24594b7a3b2c220b3669bb819f984f55310a1ae7823651d4a02e99447972595139363434e5e30a7e7d241551e1b9", 16),
                new BigInteger("07d3e47bf686600b11ac283ce88dbb3f6051e8efd04680e44c171ef531b80b2b7c39fc766320e2cf15d8d99820e96ff30dc69691839c4b40d7b06e45307dc91f3f", 16));
        AsymmetricRsaPublicKey testPublicKey = new AsymmetricRsaPublicKey(FipsRsa.Alg,
                new BigInteger("495370a1fb18543c16d3631e3163255df62be6eee890d5f25509e4f778a8ea6fbbbcdf85dff64e0d972003ab3681fbba6dd41fd541829b2e582de9f2a4a4e0a2d0900bef4753db3cee0ee06c7dfae8b1d53b5953218f9cceea695b08668edeaadced9463b1d790d5ebf27e9115b46cad4d9a2b8efab0561b0810344739ada0733f", 16),
                new BigInteger("010001", 16));


        CertificateRequestMessage crmfRequest = createCertificateRequestMessage(testPublicKey, testPrivateKey);
        byte[] crmfPubKey = crmfRequest.GetCertTemplate().PublicKey.GetEncoded();
        byte[] crmfSign   = PopoSigningKey.GetInstance(crmfRequest.ToAsn1Structure().Popo.Object).Signature.GetBytes();
        Console.WriteLine(string.Format("CRMF Request:\t\tPub Key=[{0}]\tSign=[{1}]",
            string.Join("",ExValues.SubArray(crmfPubKey,0,20)) + "..." + 
            string.Join("",ExValues.SubArray(crmfPubKey,crmfPubKey.Length-20,20)),
            string.Join("",ExValues.SubArray(crmfSign,0,20)) + "..." + 
            string.Join("",ExValues.SubArray(crmfSign,crmfSign.Length-20,20))
            ));
            
        bool result = isValidCertificateRequestMessage(crmfRequest);
        Console.WriteLine(string.Format("CRMF Verification\tverify=[{0}]",
           result ? "True" : "False"));
        Console.WriteLine();

        crmfRequest = createEncCertificateRequestMessage(testPublicKey, testPrivateKey);
        crmfPubKey = crmfRequest.GetCertTemplate().PublicKey.GetEncoded();
        Console.WriteLine(string.Format("CRMF ENC Request:\tPub Key=[{0}]",
            string.Join("",ExValues.SubArray(crmfPubKey,0,20)) + "..." + 
            string.Join("",ExValues.SubArray(crmfPubKey,crmfPubKey.Length-20,20))
            ));
        Console.WriteLine();

        
            
        Console.WriteLine("============================");
        Console.WriteLine("End   Of   CrmfExamples.Main");
        Console.WriteLine("============================");
        Console.WriteLine();
        Console.WriteLine();        

    }
    

    /// <summary>
    /// Example 49 – A Basic CRMF Request
    /// </summary>
    public static CertificateRequestMessage createCertificateRequestMessage(        
        AsymmetricRsaPublicKey pubKey, AsymmetricRsaPrivateKey priKey)
    {
        CertificateRequestMessageBuilder certReqBuilder = 
            new CertificateRequestMessageBuilder(BigInteger.One);

        ISignatureFactory<IParameters<Algorithm>> sigFact = CryptoServicesRegistrar.CreateService(
            priKey, new SecureRandom()).CreateSignatureFactory(FipsRsa.Pkcs1v15.WithDigest(FipsShs.Sha256));
        PkixSignatureFactory pkixSignFact = new PkixSignatureFactory(sigFact);

        certReqBuilder = certReqBuilder.SetSubject(new X500Name("CN=CRMF Example"));
        //certReqBuilder = certReqBuilder.SetAuthInfoSender(new X500Name("CN=CRMF Example"));
        certReqBuilder = certReqBuilder.SetPublicKey(SubjectPublicKeyInfo.GetInstance(pubKey.GetEncoded()));
        certReqBuilder = certReqBuilder.SetProofOfPossessionSignKeySigner(pkixSignFact);
        return certReqBuilder.Build();
    }
    
    public static bool isValidCertificateRequestMessage(CertificateRequestMessage certReqMsg)
    {

        IAsymmetricPublicKey pubKey = AsymmetricKeyFactory.CreatePublicKey(
            certReqMsg.GetCertTemplate().PublicKey.GetEncoded());
        PkixVerifierFactoryProvider pro = new PkixVerifierFactoryProvider(pubKey);       
        bool result = certReqMsg.IsValidSigningKeyPop(pro);
        return result;
    }
      
        

    /// <summary>
    /// Example 50 – A CRMF Request for Encryption Only Keys
    /// </summary>
    public static CertificateRequestMessage createEncCertificateRequestMessage(
        AsymmetricRsaPublicKey pubKey, AsymmetricRsaPrivateKey priKey)
    {
        CertificateRequestMessageBuilder certReqBuilder = 
            new CertificateRequestMessageBuilder(BigInteger.One);

        certReqBuilder = certReqBuilder.SetSubject(new X500Name("CN=CRMF Example"));
        certReqBuilder = certReqBuilder.SetPublicKey(SubjectPublicKeyInfo.GetInstance(pubKey.GetEncoded()));
        certReqBuilder = certReqBuilder.SetProofOfPossessionSubsequentMessage(SubsequentMessage.encrCert);
        return certReqBuilder.Build();

    }        
    


}
