using System;
using System.Text;
using System.Reflection;
using System.Runtime.InteropServices;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Crypto.Asymmetric;
using Org.BouncyCastle.Crypto.General;
using Org.BouncyCastle.Operators;
using Org.BouncyCastle.Asn1;
using Org.BouncyCastle.Asn1.Crmf;
using Org.BouncyCastle.Asn1.Pkcs;
using Org.BouncyCastle.Asn1.Nist;
using Org.BouncyCastle.Asn1.X500;
using Org.BouncyCastle.Asn1.X509;
using Org.BouncyCastle.Cert;
using Org.BouncyCastle.Crmf;
using Org.BouncyCastle.Pkcs;
using Org.BouncyCastle.OpenSsl;
using Org.BouncyCastle.Ocsp;
using Org.BouncyCastle.Pkix;
using Org.BouncyCastle.Utilities.Date;
using Org.BouncyCastle.Operators.Parameters;
using bcfipsin100.util;


public class PBEncStorage
{
    
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("============================================");
        Console.WriteLine("Start Of Password Based Encryption & Storage");
        Console.WriteLine("============================================");
    
    
        ////////////////////////////////////////////////////////////////////////
        // Example 59 – Password Based Key Generation
        ////////////////////////////////////////////////////////////////////////
        if (1 != 0)
        {
            Console.WriteLine();
            Console.WriteLine("*** Example 59 – Password Based Key Generation ***");
            byte[] result = GenerateSecretKey(ExValues.samplePassword,
                                              ExValues.sampleSalt, 2048, 256);
            Console.WriteLine();
            Console.WriteLine(string.Format("Original Password:[{0}]\t PBKDF:[{1}]",
                              new string(ExValues.samplePassword),
                              Convert.ToHexString(result)));
            Console.WriteLine();
        }
    
    

        ////////////////////////////////////////////////////////////////////////
        // Example 60 – RSA Public Key Encoding/Decoding
        ////////////////////////////////////////////////////////////////////////
        if (1 != 0)
        {
            Console.WriteLine();
            Console.WriteLine("*** Example 60 – RSA Public Key Encoding ***");
            
            AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey>  keyPair = 
                GenerateRSAKeyPair(BigInteger.ValueOf(0x10001), 2048);

            Console.WriteLine();
            Console.WriteLine(string.Format("Modulus Public Key Before Encoded:[{0}]",keyPair.PublicKey.Modulus.ToString(16)));
                
            byte[] result = EncodePublicKey(keyPair.PublicKey);
            Console.WriteLine();
            Console.WriteLine(string.Format("PKCS#1 Public Key Encoded Bytes:[{0}]",Convert.ToHexString(result)));
                      
            AsymmetricRsaPublicKey pubKey = DecodePublicKey(result);
            Console.WriteLine();
            Console.WriteLine("*** Example 60 – RSA Public Key Decoding ***");
            Console.WriteLine();
            Console.WriteLine(string.Format("PKCS#1 Public Key Decoded Modulus:[{0}]",pubKey.Modulus.ToString(16)));
            Console.WriteLine();
        }
   
   
        ////////////////////////////////////////////////////////////////////////
        // Example 61 – RSA Private Key Encoding/Decoding
        ////////////////////////////////////////////////////////////////////////
        if (1 != 0)
        {
            Console.WriteLine();
            Console.WriteLine("*** Example 61 – RSA Private Key Encoding ***");
            
            AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey>  keyPair = 
                GenerateRSAKeyPair(BigInteger.ValueOf(0x10001), 2048);
            
            Console.WriteLine();
            Console.WriteLine(string.Format("Modulus Private Key Before Encoded:[{0}]",keyPair.PrivateKey.Modulus.ToString(16)));
            
            byte[] result = EncodePrivateKey(keyPair.PrivateKey);
            Console.WriteLine();
            Console.WriteLine(string.Format("PKCS#1 Private Key Encoded:[{0}]",Convert.ToHexString(result)));
            
            
            AsymmetricRsaPrivateKey privateKey = DecodePrivateKey(result);
            Console.WriteLine();
            Console.WriteLine("*** Example 61 – RSA Private Key Encoding ***");
            Console.WriteLine();
            Console.WriteLine(string.Format("PKCS#1 Private Key Decoded Modulus:[{0}]",privateKey.Modulus.ToString(16)));
            Console.WriteLine();
        }      

        

        ////////////////////////////////////////////////////////////////////////
        // Example 62 – Reading/Writing a Certificate in PEM Format
        ////////////////////////////////////////////////////////////////////////
        if (1 != 0)
        {
            Console.WriteLine();
            Console.WriteLine("*** Example 62 – Writing a Certificate in PEM Format ***");
            string pemCert = WritePemCertificate();
            Console.WriteLine();
            Console.WriteLine(pemCert);
            
            Console.WriteLine();
            Console.WriteLine("*** Example 62 – Reading a Certificate in PEM Format ***");
            X509Certificate x509CertV1 = ReadPemCertificate(pemCert);
            Console.WriteLine();
            Console.WriteLine(x509CertV1.ToString());
        }
        

        ////////////////////////////////////////////////////////////////////////
        // Example 62 – Reading An openssl generated certificate
        ////////////////////////////////////////////////////////////////////////
        if (1 != 0)
        {
        
            Console.WriteLine();
            Console.WriteLine("*** Example 62 – Reading an openSSL generated Certificate in PEM Format ***");
            string openSSLCert = "-----BEGIN CERTIFICATE-----\n" + 
                                "MIIC/zCCAecCFEqbJgWyhPwLh8vBWrk3ZkLtQwqCMA0GCSqGSIb3DQEBCwUAMDwx\n" +
                                "CzAJBgNVBAYTAkFVMQswCQYDVQQIDAJXQTEOMAwGA1UEBwwFUGVydGgxEDAOBgNV\n" +
                                "BAoMB1dlYkNhc2gwHhcNMjQwNDE2MTcxOTE5WhcNMjUwNDE2MTcxOTE5WjA8MQsw\n" +
                                "CQYDVQQGEwJBVTELMAkGA1UECAwCV0ExDjAMBgNVBAcMBVBlcnRoMRAwDgYDVQQK\n" +
                                "DAdXZWJDYXNoMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA5jl7gRhB\n" +
                                "N59oiRskEZyrpAja08Pcrdi/EEDl6x1jxb3dP+Ber6IrAK1Xxl3/oqHzH/iGdLr/\n" +
                                "0kr7Yn0JumCVbP0iaEG8GquVhaA3crK44/lOnungwH9JVHC4DmSgeZVgldKximUD\n" +
                                "7+67kRxl8BtnL2KtsvbiFzb0HxPnm/rHBtARuKUlAEKIBqV7k2ku6GHQy5e6XPj/\n" +
                                "/JEZNJKcXUt7gDqt2CYl4fuKkKF0Pf69/T/1UAQEQGeGBjcjJbsO6bDBwcToKIjF\n" +
                                "fm0b0u6H87AGUL2LwHx/39p51V+t/GAtd1FMRv1cQwqM+xx+XqCftJ3nR87Co4EE\n" +
                                "akNIADn15bbs2wIDAQABMA0GCSqGSIb3DQEBCwUAA4IBAQCGpgOavamroXntrC1C\n" +
                                "eJgsS+iMc/Idcv/7G7zng/GNQgIQk743FY0qePGW+k1lesXU5uY2oy4EVPrm/HX4\n" +
                                "R7M24/7RAEwiBr3ba+/xNvbxWyMWhgDFVC0UZT5kxpz+GLL+tpibBAOU2THT36h4\n" +
                                "qrBiqjFFBfIZxF1SpYSTPob8rpZCJXOAb572n2qi5aZVYX3C3tBpkkIVq6xE6iwx\n" +
                                "MCe1nX6kJQm81ocx//PqgqZrLxSjffe81Wjjiig6dyjY6Bfb+20TAQarp47gq8yj\n" +
                                "hdg87pXyQvVrVkbSZPGRfrULo40M1R+EvnTnJIWyyh6AQCijK6v/aI7hI7qiYuNy\n" +
                                "gvzm\n" +
                                "-----END CERTIFICATE-----";
            X509Certificate x509CertV1 = ReadPemCertificate(openSSLCert);
            Console.WriteLine();
            Console.WriteLine(x509CertV1.ToString());
        }


        ////////////////////////////////////////////////////////////////////////
        // Example 63 – Reading/Writing a Private Key in PEM Format
        ////////////////////////////////////////////////////////////////////////
        if (1 != 0)
        {
            AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey>  keyPair = 
                GenerateRSAKeyPair(BigInteger.ValueOf(0x10001), 2048);

            Console.WriteLine();
            Console.WriteLine("*** Example 63 – Writing a Private Key in PEM Format ***");
            
            Console.WriteLine();
            Console.WriteLine(string.Format("Modulus Private Key Before Encoded:[{0}]",keyPair.PrivateKey.Modulus.ToString(16)));
            string pemKey = WritePrivateKey(keyPair.PrivateKey);
            Console.WriteLine();
            Console.WriteLine(pemKey);
            
            Console.WriteLine();
            Console.WriteLine("*** Example 63 – Reading a Private Key in PEM Format ***");
            AsymmetricRsaPrivateKey rsaKey = ReadPrivateKey(pemKey);
            Console.WriteLine();
            Console.WriteLine(string.Format("Modulus Private Key After Decoded:[{0}]",rsaKey.Modulus.ToString(16)));
        }


        ////////////////////////////////////////////////////////////////////////
        // Example 64 – Reading/Writing an Encrypted Private Key in PEM Format
        ////////////////////////////////////////////////////////////////////////
        if (1 != 0)
        {
            AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey>  keyPair = 
                GenerateRSAKeyPair(BigInteger.ValueOf(0x10001), 2048);

            Console.WriteLine();
            Console.WriteLine("*** Example 64 – Writing an Encrypted Private Key in PEM Format ***");

            Console.WriteLine();
            Console.WriteLine(string.Format("Modulus Private Key Before Encoded:[{0}]",keyPair.PrivateKey.Modulus.ToString(16)));
            string encryptedPemKey = WriteEncryptedKey(ExValues.simplePassword, keyPair.PrivateKey);            
            Console.WriteLine();
            Console.WriteLine(encryptedPemKey);
            
            Console.WriteLine();
            Console.WriteLine("*** Example 64 – Reading an Encrypted Private Key in PEM Format ***");
            AsymmetricRsaPrivateKey rsaKey = ReadEncryptedKey(ExValues.simplePassword,encryptedPemKey);
            Console.WriteLine();
            Console.WriteLine(string.Format("Modulus Private Key After Decoded:[{0}]",rsaKey.Modulus.ToString(16)));
        
        }


        ////////////////////////////////////////////////////////////////////////
        // Example 65 – Reading/Writing Encrypted Private Key (OpenSSL Style)
        ////////////////////////////////////////////////////////////////////////
        if (1 != 0)
        {
            AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey>  keyPair = 
                GenerateRSAKeyPair(BigInteger.ValueOf(0x10001), 2048);

            Console.WriteLine();
            Console.WriteLine("*** Example 65 – Writing Encrypted Private Key (OpenSSL Style) ***");

            Console.WriteLine();
            Console.WriteLine(string.Format("Modulus Private Key Before Encoded:[{0}]",keyPair.PrivateKey.Modulus.ToString(16)));
            string encryptedPemKey = WriteEncryptedKeyOpenSSL(ExValues.simplePassword, keyPair.PrivateKey);            
            Console.WriteLine();
            Console.WriteLine(encryptedPemKey);
            
            Console.WriteLine();
            Console.WriteLine("*** Example 65 – Reading Encrypted Private Key (OpenSSL Style) ***");
            AsymmetricRsaPrivateKey rsaKey = ReadEncryptedKeyOpenSSL(ExValues.simplePassword,encryptedPemKey);
            Console.WriteLine();
            Console.WriteLine(string.Format("Modulus Private Key After Decoded:[{0}]",rsaKey.Modulus.ToString(16)));
        
        }
        
        Console.WriteLine("============================================");
        Console.WriteLine("End   Of Password Based Encryption & Storage");
        Console.WriteLine("============================================");
        Console.WriteLine();
        Console.WriteLine();        
    }
    
   
    /// <summary>
    /// Example 59 – Password Based Key Generation
    /// </summary>
    public static byte[] GenerateSecretKey(char[] password, byte[] salt, 
                                           int iterationCount = 2048,
                                           int keySizeInBits = 256)
    {


        IPasswordBasedDeriverBuilder<FipsPbkd.Parameters> pbkdfBuilder =
            CryptoServicesRegistrar.CreateService(FipsPbkd.PbkdF2)
                .From(PasswordConverter.UTF8, password);

        IPasswordBasedDeriver<FipsPbkd.Parameters> pbkdfDeriver = 
            pbkdfBuilder.WithPrf(FipsShs.Sha384)
                        .WithSalt(salt)
                        .WithIterationCount(iterationCount)
                        .Build();
                             
        return pbkdfDeriver.DeriveKey(TargetKeyType.CIPHER, (keySizeInBits / 8));
    }



    /// <summary>
    /// Example 60 – RSA Public Key Encoding
    /// </summary>
    public static byte[] EncodePublicKey(IAsymmetricPublicKey publicKey)
    {
        return publicKey.GetEncoded();
    }

    /// <summary>
    /// Example 60 – RSA Public Key Decoding
    /// </summary>
    public static  AsymmetricRsaPublicKey DecodePublicKey(byte[] publicKey)
    {
        return new AsymmetricRsaPublicKey(FipsRsa.Alg, publicKey);
    }



    /// <summary>
    /// Example 61 – RSA Private Key Encoding
    /// </summary>
    public static byte[] EncodePrivateKey(IAsymmetricPrivateKey privateKey)
    {
        return privateKey.GetEncoded();
    }

    /// <summary>
    /// Example 61 – RSA Private Key Decoding
    /// </summary>
    public static  AsymmetricRsaPrivateKey DecodePrivateKey(byte[] privateKey)
    {
        return new AsymmetricRsaPrivateKey(FipsRsa.Alg, privateKey);
    }

    
    
    /// <summary>
    /// Example 62 – Writing a Certificate in PEM Format
    /// </summary>
    public static string WritePemCertificate()
    {
        AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  generateKeyPair()
        {
            FipsEC.KeyGenerationParameters keyGenParametersCA =
                new FipsEC.KeyGenerationParameters(FipsEC.DomainParams.K409).For(FipsEC.Cdh);
            ECDomainParameters ecParamsCA = keyGenParametersCA.DomainParameters;    
            FipsEC.KeyPairGenerator kpGenCA = 
                CryptoServicesRegistrar.CreateGenerator(keyGenParametersCA, ExValues.DHFixedRandomA(1000));
            return kpGenCA.GenerateKeyPair();
        }
           
        
        AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPairEC = generateKeyPair();
            

                
        X509Certificate makeX509V3Certificate()
        {
        
            X509V1CertificateGenerator certGen = new X509V1CertificateGenerator();
            BigInteger serialNumber = ExValues.FixedRandomBigInteger(32,1);
            certGen.SetSerialNumber(serialNumber);
            certGen.SetIssuerDN(new X500Name("CN=Issuer Self_Signed_CA"));
            certGen.SetNotBefore(DateTime.UtcNow.AddSeconds(-50));
            certGen.SetNotAfter(DateTime.UtcNow.AddSeconds(50));    
            certGen.SetSubjectDN(new X500Name("CN=Subject Self_Signed_CA"));
            certGen.SetPublicKey(keyPairEC.PublicKey);

            ISignatureFactoryService signatureFactoryProvider =
                CryptoServicesRegistrar.CreateService(keyPairEC.PrivateKey, ExValues.DHFixedRandomA(1000));

            ISignatureFactory<FipsEC.SignatureParameters> signer = 
                signatureFactoryProvider.CreateSignatureFactory(FipsEC.Dsa);
                
            return certGen.Generate(new PkixSignatureFactory(signer));
        }
        
        X509Certificate  x509CertEE = makeX509V3Certificate();
        
        StringWriter strWriterEE = new StringWriter();
        OpenSslPemWriter pemWriterEE = new OpenSslPemWriter(strWriterEE);
        pemWriterEE.WriteObject(x509CertEE);
        return strWriterEE.ToString();
    }

    /// <summary>
    /// Example 62 – Reading a Certificate in PEM Format
    /// </summary>
    public static X509Certificate ReadPemCertificate(string pemCert)
    {
        StringReader strReader = new StringReader(pemCert);
        OpenSslPemReader pemReader = new OpenSslPemReader(strReader);
                  
        X509Certificate x509Cert = (X509Certificate)pemReader.ReadObject();
        return x509Cert;
    }
    
 
    /// <summary>
    /// Example 63 – Writing a Private Key in PEM Format
    /// </summary>
    public static string WritePrivateKey(AsymmetricRsaPrivateKey privateKey)
    {
        byte[] keyEncoding = privateKey.GetEncoded();
        PrivateKeyInfo keyInfo = PrivateKeyInfo.GetInstance(keyEncoding);
        StringWriter strWriter = new StringWriter();
        OpenSslPemWriter pemWriter = new OpenSslPemWriter(strWriter);
        pemWriter.WriteObject(keyInfo);
        return strWriter.ToString();
    }    

    /// <summary>
    /// Example 63 – Reading a Private Key in PEM Format
    /// </summary>
    public static AsymmetricRsaPrivateKey ReadPrivateKey(string key)
    {
        StringReader strReader = new StringReader(key);
        OpenSslPemReader pemReader = new OpenSslPemReader(strReader);
        PemKeyPair keyInfo  = (PemKeyPair)pemReader.ReadObject();
        PrivateKeyInfo privateKeyInfo  = keyInfo.PrivateKeyInfo;        
        byte[] privateKeyData = privateKeyInfo.PrivateKeyData.GetOctets();
        AsymmetricRsaPrivateKey privateKey = new AsymmetricRsaPrivateKey(FipsRsa.Alg, privateKeyData);       
        return privateKey;
    }    


    /// <summary>
    /// Example 64 – Writing an Encrypted Private Key in PEM Format
    /// </summary>
    public static string WriteEncryptedKey(char[] passwd, IAsymmetricPrivateKey privateKey)
    {
        byte[] asn1KeyInfo = privateKey.GetEncoded();        
        Pkcs8EncryptedPrivateKeyInfoBuilder pkcs8Builder =  new Pkcs8EncryptedPrivateKeyInfoBuilder(asn1KeyInfo);

        PkixPbeEncryptorBuilder pbeEncryptBuilder = new PkixPbeEncryptorBuilder(NistObjectIdentifiers.IdAes256Ccm);
        pbeEncryptBuilder = pbeEncryptBuilder.WithSecureRandom(ExValues.DHFixedRandomA(60000)).WithSalt(ExValues.sampleSalt);
        ICipherBuilder<AlgorithmIdentifier> pbecipherBuilder = pbeEncryptBuilder.Build(passwd); 
        
        Pkcs8EncryptedPrivateKeyInfo encryptedKeyInfo = pkcs8Builder.Build(pbecipherBuilder);
        
        StringWriter strWriter = new StringWriter();
        OpenSslPemWriter pemWriter = new OpenSslPemWriter(strWriter);
        pemWriter.WriteObject(encryptedKeyInfo);
        return strWriter.ToString();    
    }
    
    /// <summary>
    /// Example 64 – Writing an Encrypted Private Key in PEM Format
    /// </summary>
    public static AsymmetricRsaPrivateKey ReadEncryptedKey(char[] passwd, string pemKey)
    {
        
        StringReader strReader = new StringReader(pemKey);
        OpenSslPemReader pemReader = new OpenSslPemReader(strReader);
        Pkcs8EncryptedPrivateKeyInfo encryptedKeyInfo  = (Pkcs8EncryptedPrivateKeyInfo)pemReader.ReadObject();
        
        PrivateKeyInfo privateKeyInfo = encryptedKeyInfo.DecryptPrivateKeyInfo(new PkixPbeDecryptorProviderBuilder().Build(passwd));
        
        byte[] privateKeyData = privateKeyInfo.PrivateKeyData.GetOctets();
        AsymmetricRsaPrivateKey privateKey = new AsymmetricRsaPrivateKey(FipsRsa.Alg, privateKeyData);       
        return privateKey;
    }
    
    
    /// <summary>
    /// Example 65 – Writing Encrypted Private Key (OpenSSL Style)
    /// </summary>
    public static string WriteEncryptedKeyOpenSSL(char[] passwd, IAsymmetricPrivateKey privateKey)
    {
        PrivateKeyInfo keyInfo = PrivateKeyInfo.GetInstance(privateKey.GetEncoded());
        OpenSslPemEncryptorBuilder encryptBuilder = new OpenSslPemEncryptorBuilder(DekAlgorithm.Aes256Cbc);
        ICipherBuilder<DekInfo> cipherBuilder = encryptBuilder.WithSecureRandom(ExValues.DHFixedRandomA(60000)).Build(passwd);
        
        StringWriter strWriter = new StringWriter();
        OpenSslPemWriter pemWriter = new OpenSslPemWriter(strWriter);
        pemWriter.WriteObject(keyInfo, cipherBuilder);
        return strWriter.ToString();    
    }    
    
    /// <summary>
    /// Example 65 – Reading Encrypted Private Key (OpenSSL Style)
    /// </summary>
    public static AsymmetricRsaPrivateKey ReadEncryptedKeyOpenSSL(char[] passwd, string pemKey)
    {
        
        StringReader strReader = new StringReader(pemKey);
        OpenSslPemReader pemReader = new OpenSslPemReader(strReader);
        
        OpenSslEncryptedObject encryptedKeyInfo  = (OpenSslEncryptedObject)pemReader.ReadObject();
        
        PemKeyPair keyInfo = (PemKeyPair)encryptedKeyInfo.Decrypt(new OpenSslPemDecryptorBuilderProvider(passwd));        
        PrivateKeyInfo privateKeyInfo  = keyInfo.PrivateKeyInfo;        
        byte[] privateKeyData = privateKeyInfo.PrivateKeyData.GetOctets();
        AsymmetricRsaPrivateKey privateKey = new AsymmetricRsaPrivateKey(FipsRsa.Alg, privateKeyData);       
        return privateKey;
    }
    
    
    public static AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey> 
        GenerateRSAKeyPair(BigInteger e, int keySize)  
    {   

        FipsRsa.KeyGenerationParameters keyGenParameters =
            new FipsRsa.KeyGenerationParameters(e, keySize);
                
        FipsRsa.KeyPairGenerator kpGen = 
            CryptoServicesRegistrar.CreateGenerator(keyGenParameters, ExValues.DHFixedRandomA(60000));

        return kpGen.GenerateKeyPair();
    }
    

}
