using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Crypto.Asymmetric;
using Org.BouncyCastle.Crypto.Internal.Prng;

namespace bcfipsin100.util
{

    public class ExValues
    {
        public static readonly long thirty_days = 1000L * 60 * 60 * 24 * 30;
        public static readonly byte[] sampleAesKeyInput = Hex.Decode("000102030405060708090a0b0c0d0e0f");
        public static readonly FipsAes.Key sampleAesKey = new FipsAes.Key(sampleAesKeyInput);    

        public static readonly byte[] sampleTripleDesKeyInput =
                    Hex.Decode("000102030405060708090a0b0c0d0e0f1011121314151617");
                    
        public static readonly SymmetricSecretKey sampleTripleDesKey = 
                    new FipsTripleDes.Key(sampleTripleDesKeyInput);
           
        public static readonly SymmetricSecretKey sampleHMacKey = 
                    new FipsShs.Key(FipsShs.Sha512HMac,Hex.Decode("000102030405060708090a0b0c0d0e0f10111213"));

        public static readonly byte[] sampleIVnonce = Strings.ToByteArray("0123456789123456");
        public static readonly byte[] sampleIVcounter = Strings.ToByteArray("012345678912");
        public static readonly byte[] sampleInput = Strings.ToByteArray("Hello World!");
        public static readonly byte[] sampleNonce = Strings.ToByteArray("number only used once");
        public static readonly char[] samplePassword = "My PBKDF Password is \"Hello World!\"".ToCharArray();
        public static readonly char[] simplePassword = "HiThere".ToCharArray();
        public static readonly byte[] sampleSalt = Strings.ToByteArray("My PBKDF Salt To Use");

        public static readonly byte[] sampleTwoBlockInput =
                    Strings.ToByteArray("Some cipher modes require more than one block");

        public static readonly byte[] personalizationString = 
                    Strings.ToByteArray("a constant personal marker");

        public static readonly byte[] additionalAuthenticatedDataA = 
                        Strings.ToByteArray("This message was sent 29th Feb at 11.00am - does not repeat");
                        
        public static readonly byte[] additionalAuthenticatedDataB = 
                        Strings.ToByteArray("This message was sent 30th Feb at 11.00am - does not repeat");

        public static readonly byte[] sampleSignatureMessage =
                        Strings.ToByteArray("Once Upon a midnight dreary, while I pondered weak and weary");        

        public static readonly byte[] initiator = Strings.ToByteArray("Initiator");
        public static readonly byte[] recipient = Strings.ToByteArray("Recipient");
        public static readonly byte[] sampleUKM = Strings.ToByteArray("User keying material");

        // sample byte array multiple of 8 bytes - used for demonstrating key wrapping block aligned no padding
        public static readonly byte[] sampleUnpaddedKey = Strings.ToByteArray("this is my secret key123");
        // sample byte array which is not a multiple of 8 bytes - used for demonstraing padded wrapping
        public static readonly byte[] samplePaddedKey = Strings.ToByteArray("this is my secret key123456");
        
        public static readonly BigInteger dhP = 
            new BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc32" +
                           "6acccb89b897cf348c6331881d7890675a67acf5" +
                           "f82c9c196924fd50db59f138602e065a018c4521" + 
                           "3ec88c355d2a419b75cd32e186132cbda13da9df" + 
                           "23cded3b3aa9cdda2127f98226feeb055dfeb549" + 
                           "c6e7beb14d1c8a2b30b81377185468a0c05c6aa3" + 
                           "410fd746ce38691d92cd08c5e08b959a16f5b33d" + 
                           "3a6ec71af88397cac8ed3296ac19b98fccac57a3" + 
                           "864b3a6f20873cd09f22aa66d049c20729f5e99d" + 
                           "3be168cd4a0438072f868305af5452537cc462b5" + 
                           "e6d9a65f9e19909cad2b37e3ad018dc954ab65b9" + 
                           "1b5c42551e5fe8bad45a0b555c0afd533f98f0b1" + 
                           "57e07f24dc2b9a6b465e05b029e79af1", 16);
        public static readonly BigInteger dhQ = 
            new BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1", 16);
        public static readonly BigInteger dhG = 
            new BigInteger("13cfe47aa686e1c63d029ffd6327bfa543c7fe94" +
                           "8f5dc6e64c0eaeeba521fa1e01ab56401d438b04" + 
                           "06ecd134f53aaa6a64fbfa687842c6c38b973a8e" + 
                           "da72b3e132f38a9aef65e368b434e2c5383baac8" +
                           "efe4d5635097cd74d6b0594c3509217da5990d75" + 
                           "fe4a15709992a155a7ecbbc1bae6927a6c357e90" + 
                           "910265d8902d864abd7b0a1e2cce2f42da88639a" + 
                           "9329fe1faff8cc19938de37a94226fb63a3f8c1f" + 
                           "b3fead755d181df4711322ae006ddc0956855ba6" + 
                           "53d7f8f8636d3c3c6a2b146d321a634245b20eb4" + 
                           "56e044a8fbd3d7a96e096b51a23aee1c8c162d91" + 
                           "670b9d634126e626a93c3a2fc27f0d7e7865dd02" +
                           "56b12068d20f2e2f975d74e2ccbd42ac", 16);
        
        public static byte[] SubArray(byte[] sourceArray, int sourceStart, int length) 
        {
            byte[] result = new byte[length];
            Array.Copy(sourceArray,sourceStart,result,0,length);
            return result;
        }
    
        public static readonly int fixedRandomSeed = 151;
        public static Random fixedRandom = new Random(fixedRandomSeed);
        
        public static byte[] FixedRandomBytes(int length)
        {
            byte[] result = new byte[length];
            fixedRandom.NextBytes(result);
            return result;
        }

        public static BigInteger FixedRandomBigInteger(int length, int sign=0)
        {            
            byte[] result = new byte[length];
            fixedRandom.NextBytes(result);
            if (sign == 0) {
                return new BigInteger(result);
            }
            return new BigInteger(sign,result);
        }
        
        public static SecureRandom DHFixedRandomA(int length)
        {
            int iterations = (int)(length/64) + 1;
            FixedSecureRandom.Source[] bigData = new FixedSecureRandom.Source[iterations];
            
            for (int j = 0; j < iterations; j++)
            {
                bigData[j] = (FixedSecureRandom.Source)(new FixedSecureRandom.Data(FixedRandomBytes(64)));
            }
            
            return new FixedSecureRandom(bigData);
        }
        

                
        public static readonly SecureRandom dhFixedRandom = new FixedSecureRandom(
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("13cfe47aa686e1c63d029ffd6327bfa543c7fe948f5dc6e64c0eaeeb"),
            new FixedSecureRandom.BigInteger("13cfe47aa686e1c63d029ffd6327bfa543c7fe948f5dc6e64c0eaeeb"),
            new FixedSecureRandom.BigInteger("13cfe47aa686e1c63d029ffd6327bfa543c7fe948f5dc6e64c0eaeeb"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("13cfe47aa686e1c63d029ffd6327bfa543c7fe948f5dc6e64c0eaeeb"),
            new FixedSecureRandom.BigInteger("13cfe47aa686e1c63d029ffd6327bfa543c7fe948f5dc6e64c0eaeeb"),
            new FixedSecureRandom.BigInteger("13cfe47aa686e1c63d029ffd6327bfa543c7fe948f5dc6e64c0eaeeb"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("13cfe47aa686e1c63d029ffd6327bfa543c7fe948f5dc6e64c0eaeeb"),
            new FixedSecureRandom.BigInteger("13cfe47aa686e1c63d029ffd6327bfa543c7fe948f5dc6e64c0eaeeb"),
            new FixedSecureRandom.BigInteger("13cfe47aa686e1c63d029ffd6327bfa543c7fe948f5dc6e64c0eaeeb"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("a521fa1e01ab56401d438b0406ecd134f53aaa6a64fbfa687842c6c3"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("8b973a8eda72b3e132f38a9aef65e368b434e2c5383baac8efe4d563"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("cf3c1ad8120c10f03c63383a841f0f34c28e0c968f7aad9422715df1"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34"),
            new FixedSecureRandom.BigInteger("b1793382ef28b54fb6715857c556dc8ced00dc326acccb89b897cf34")
            );
            
        public static readonly SecureRandom dhKeyAFixedRandom = new FixedSecureRandom(
            new FixedSecureRandom.BigInteger("a1b2c3213cfe47aa686e1c63d029ffd6327bfa543c7fe94a"), 
            new FixedSecureRandom.BigInteger("010203a8f5dc6e64c0eaeeba521fa1e01ab56401d438b04b"), 
            new FixedSecureRandom.BigInteger("909190256e044a8fbd3d7a96e096b51a23aee1c8c162d91b"), 
            new FixedSecureRandom.BigInteger("111213406ecd134f53aaa6a64fbfa687842c6c38b973a8ec"), 
            new FixedSecureRandom.BigInteger("212234da72b3e132f38a9aef65e368b434e2c5383baac8dd"),
            new FixedSecureRandom.BigInteger("3132334efe4d5635097cd74d6b0594c3509217da5990d75e"), 
            new FixedSecureRandom.BigInteger("909190256e044a8fbd3d7a96e096b51a23aee1c8c162d91b"), 
            new FixedSecureRandom.BigInteger("afadac5fe4a15709992a155a7ecbbc1bae6927a6c357e90f"), 
            new FixedSecureRandom.BigInteger("6563616910265d8902d864abd7b0a1e2cce2f42da88639af"), 
            new FixedSecureRandom.BigInteger("71787389329fe1faff8cc19938de37a94226fb63a3f8c1fe"), 
            new FixedSecureRandom.BigInteger("4145425b3fead755d181df4711322ae006ddc0956855ba6d"), 
            new FixedSecureRandom.BigInteger("515652753d7f8f8636d3c3c6a2b146d321a634245b20eb4c"), 
            new FixedSecureRandom.BigInteger("909190256e044a8fbd3d7a96e096b51a23aee1c8c162d91b"), 
            new FixedSecureRandom.BigInteger("5160420670b9d634126e626a93c3a2fc27f0d7e7865dd02a"),
            new FixedSecureRandom.BigInteger("909190256e044a8fbd3d7a96e096b51a23aee1c8c162d91b"), 
            new FixedSecureRandom.BigInteger("109283756b12068d20f2e2f975d74e2ccbd42acc2c31213a")
            );

        public static readonly SecureRandom dhKeyBFixedRandom = new FixedSecureRandom(
            new FixedSecureRandom.BigInteger("010203a8f5dc6e64c0eaeeba521fa1e01ab56401d438b04b"), 
            new FixedSecureRandom.BigInteger("a1b2c3213cfe47aa686e1c63d029ffd6327bfa543c7fe94a"), 
            new FixedSecureRandom.BigInteger("212234da72b3e132f38a9aef65e368b434e2c5383baac8dd"),
            new FixedSecureRandom.BigInteger("5160420670b9d634126e626a93c3a2fc27f0d7e7865dd02a"),
            new FixedSecureRandom.BigInteger("111213406ecd134f53aaa6a64fbfa687842c6c38b973a8ec"), 
            new FixedSecureRandom.BigInteger("afadac5fe4a15709992a155a7ecbbc1bae6927a6c357e90f"), 
            new FixedSecureRandom.BigInteger("3132334efe4d5635097cd74d6b0594c3509217da5990d75e"), 
            new FixedSecureRandom.BigInteger("71787389329fe1faff8cc19938de37a94226fb63a3f8c1fe"), 
            new FixedSecureRandom.BigInteger("515652753d7f8f8636d3c3c6a2b146d321a634245b20eb4c"), 
            new FixedSecureRandom.BigInteger("6563616910265d8902d864abd7b0a1e2cce2f42da88639af"), 
            new FixedSecureRandom.BigInteger("909190256e044a8fbd3d7a96e096b51a23aee1c8c162d91b"), 
            new FixedSecureRandom.BigInteger("109283756b12068d20f2e2f975d74e2ccbd42acc2c31213a"),
            new FixedSecureRandom.BigInteger("515652753d7f8f8636d3c3c6a2b146d321a634245b20eb4c"), 
            new FixedSecureRandom.BigInteger("4145425b3fead755d181df4711322ae006ddc0956855ba6d"), 
            new FixedSecureRandom.BigInteger("5160420670b9d634126e626a93c3a2fc27f0d7e7865dd02a"),
            new FixedSecureRandom.BigInteger("109283756b12068d20f2e2f975d74e2ccbd42acc2c31213a")
            );

        public static AsymmetricKeyPair<AsymmetricRsaPublicKey, AsymmetricRsaPrivateKey> generateRSAKeyPair()  
        {   
            SecureRandom secRandom = DHFixedRandomA(50000);
            FipsRsa.KeyGenerationParameters keyGenParameters =
                new FipsRsa.KeyGenerationParameters(BigInteger.ValueOf(0x10001), 2048);                
            FipsRsa.KeyPairGenerator kpGen = 
                CryptoServicesRegistrar.CreateGenerator(keyGenParameters, secRandom);
            return kpGen.GenerateKeyPair();
        }      
    
    }
}